"""
Tests for the Condition class.
"""

import pytest
from datetime import datetime, timedelta

from sortmeout.core.condition import (
    Condition,
    ConditionGroup,
    ConditionGroupMode,
    ConditionOperator,
    ConditionAttribute,
    parse_size,
    parse_duration,
    name_equals,
    name_contains,
    extension_is,
    extension_in,
    size_greater_than,
    size_less_than,
    modified_within,
)


class TestParseSize:
    """Tests for size parsing."""

    def test_parse_bytes(self):
        assert parse_size("1024") == 1024
        assert parse_size("1024B") == 1024

    def test_parse_kilobytes(self):
        assert parse_size("1KB") == 1024
        assert parse_size("1K") == 1024
        assert parse_size("10KB") == 10240

    def test_parse_megabytes(self):
        assert parse_size("1MB") == 1024 ** 2
        assert parse_size("1M") == 1024 ** 2
        assert parse_size("5MB") == 5 * 1024 ** 2

    def test_parse_gigabytes(self):
        assert parse_size("1GB") == 1024 ** 3
        assert parse_size("2GB") == 2 * 1024 ** 3

    def test_parse_with_spaces(self):
        assert parse_size(" 10 MB ") == 10 * 1024 ** 2

    def test_parse_float(self):
        assert parse_size("1.5MB") == int(1.5 * 1024 ** 2)


class TestParseDuration:
    """Tests for duration parsing."""

    def test_parse_days(self):
        assert parse_duration("7 days") == timedelta(days=7)
        assert parse_duration("7d") == timedelta(days=7)

    def test_parse_hours(self):
        assert parse_duration("24 hours") == timedelta(hours=24)
        assert parse_duration("24h") == timedelta(hours=24)

    def test_parse_minutes(self):
        assert parse_duration("30 minutes") == timedelta(minutes=30)
        assert parse_duration("30m") == timedelta(minutes=30)

    def test_parse_weeks(self):
        assert parse_duration("2 weeks") == timedelta(weeks=2)
        assert parse_duration("2w") == timedelta(weeks=2)


class TestConditionEquality:
    """Tests for equality-based conditions."""

    def test_equals_case_insensitive(self):
        cond = Condition("name", "equals", "Test")
        assert cond.evaluate({"name": "test"}) is True
        assert cond.evaluate({"name": "TEST"}) is True
        assert cond.evaluate({"name": "other"}) is False

    def test_equals_case_sensitive(self):
        cond = Condition("name", "equals", "Test", case_sensitive=True)
        assert cond.evaluate({"name": "Test"}) is True
        assert cond.evaluate({"name": "test"}) is False

    def test_not_equals(self):
        cond = Condition("name", "not_equals", "Test")
        assert cond.evaluate({"name": "other"}) is True
        assert cond.evaluate({"name": "test"}) is False


class TestConditionContains:
    """Tests for contains-based conditions."""

    def test_contains(self):
        cond = Condition("name", "contains", "report")
        assert cond.evaluate({"name": "monthly_report_2024"}) is True
        assert cond.evaluate({"name": "invoice_2024"}) is False

    def test_not_contains(self):
        cond = Condition("name", "not_contains", "temp")
        assert cond.evaluate({"name": "document.pdf"}) is True
        assert cond.evaluate({"name": "temp_file.txt"}) is False

    def test_starts_with(self):
        cond = Condition("name", "starts_with", "Invoice")
        assert cond.evaluate({"name": "Invoice_2024"}) is True
        assert cond.evaluate({"name": "2024_Invoice"}) is False

    def test_ends_with(self):
        cond = Condition("name", "ends_with", "_final")
        assert cond.evaluate({"name": "document_final"}) is True
        assert cond.evaluate({"name": "final_document"}) is False


class TestConditionPatternMatching:
    """Tests for pattern matching conditions."""

    def test_matches_regex(self):
        cond = Condition("name", "matches_regex", r"^invoice_\d{4}$")
        assert cond.evaluate({"name": "invoice_2024"}) is True
        assert cond.evaluate({"name": "invoice_abc"}) is False

    def test_matches_glob(self):
        cond = Condition("name", "matches_glob", "report_*.pdf")
        assert cond.evaluate({"name": "report_2024.pdf"}) is True
        assert cond.evaluate({"name": "report_january.pdf"}) is True
        assert cond.evaluate({"name": "invoice_2024.pdf"}) is False


class TestConditionNumeric:
    """Tests for numeric conditions."""

    def test_greater_than(self):
        cond = Condition("size", "greater_than", 1024)
        assert cond.evaluate({"size": 2048}) is True
        assert cond.evaluate({"size": 512}) is False

    def test_greater_than_with_size_string(self):
        cond = Condition("size", "greater_than", "1MB")
        assert cond.evaluate({"size": 2 * 1024 * 1024}) is True
        assert cond.evaluate({"size": 512 * 1024}) is False

    def test_less_than(self):
        cond = Condition("size", "less_than", "10MB")
        assert cond.evaluate({"size": 5 * 1024 * 1024}) is True
        assert cond.evaluate({"size": 20 * 1024 * 1024}) is False

    def test_between(self):
        cond = Condition("size", "between", [1024, 4096])
        assert cond.evaluate({"size": 2048}) is True
        assert cond.evaluate({"size": 512}) is False
        assert cond.evaluate({"size": 8192}) is False


class TestConditionDate:
    """Tests for date-based conditions."""

    def test_is_today(self):
        cond = Condition("date_modified", "is_today", None)
        assert cond.evaluate({"date_modified": datetime.now()}) is True
        assert cond.evaluate({"date_modified": datetime.now() - timedelta(days=1)}) is False

    def test_is_yesterday(self):
        cond = Condition("date_modified", "is_yesterday", None)
        assert cond.evaluate({"date_modified": datetime.now() - timedelta(days=1)}) is True
        assert cond.evaluate({"date_modified": datetime.now()}) is False

    def test_within_last(self):
        cond = Condition("date_modified", "within_last", "7 days")
        assert cond.evaluate({"date_modified": datetime.now() - timedelta(days=3)}) is True
        assert cond.evaluate({"date_modified": datetime.now() - timedelta(days=10)}) is False

    def test_not_within_last(self):
        cond = Condition("date_modified", "not_within_last", "30 days")
        assert cond.evaluate({"date_modified": datetime.now() - timedelta(days=60)}) is True
        assert cond.evaluate({"date_modified": datetime.now() - timedelta(days=7)}) is False


class TestConditionList:
    """Tests for list-based conditions."""

    def test_in_list(self):
        cond = Condition("extension", "in_list", ["pdf", "doc", "docx"])
        assert cond.evaluate({"extension": "pdf"}) is True
        assert cond.evaluate({"extension": "txt"}) is False

    def test_not_in_list(self):
        cond = Condition("extension", "not_in_list", ["tmp", "temp", "bak"])
        assert cond.evaluate({"extension": "pdf"}) is True
        assert cond.evaluate({"extension": "tmp"}) is False


class TestConditionExistence:
    """Tests for existence conditions."""

    def test_exists(self):
        cond = Condition("tags", "exists", None)
        assert cond.evaluate({"tags": ["work"]}) is True
        assert cond.evaluate({"tags": None}) is False
        assert cond.evaluate({}) is False

    def test_not_exists(self):
        cond = Condition("tags", "not_exists", None)
        assert cond.evaluate({}) is True
        assert cond.evaluate({"tags": None}) is True
        assert cond.evaluate({"tags": ["work"]}) is False

    def test_is_empty(self):
        cond = Condition("tags", "is_empty", None)
        assert cond.evaluate({"tags": []}) is True
        assert cond.evaluate({"tags": ""}) is True
        assert cond.evaluate({"tags": ["work"]}) is False

    def test_is_not_empty(self):
        cond = Condition("tags", "is_not_empty", None)
        assert cond.evaluate({"tags": ["work"]}) is True
        assert cond.evaluate({"tags": []}) is False


class TestConditionNegation:
    """Tests for negated conditions."""

    def test_negated_equals(self):
        cond = Condition("extension", "equals", "pdf", negate=True)
        assert cond.evaluate({"extension": "pdf"}) is False
        assert cond.evaluate({"extension": "doc"}) is True

    def test_negated_contains(self):
        cond = Condition("name", "contains", "temp", negate=True)
        assert cond.evaluate({"name": "temp_file"}) is False
        assert cond.evaluate({"name": "document"}) is True


class TestConditionGroup:
    """Tests for condition groups."""

    def test_all_mode(self):
        group = ConditionGroup(
            conditions=[
                Condition("extension", "equals", "pdf"),
                Condition("size", "greater_than", "1MB"),
            ],
            mode=ConditionGroupMode.ALL,
        )

        # Both conditions match
        assert group.evaluate({"extension": "pdf", "size": 2 * 1024 * 1024}) is True

        # Only one condition matches
        assert group.evaluate({"extension": "pdf", "size": 512 * 1024}) is False

    def test_any_mode(self):
        group = ConditionGroup(
            conditions=[
                Condition("extension", "equals", "pdf"),
                Condition("extension", "equals", "doc"),
            ],
            mode=ConditionGroupMode.ANY,
        )

        assert group.evaluate({"extension": "pdf"}) is True
        assert group.evaluate({"extension": "doc"}) is True
        assert group.evaluate({"extension": "txt"}) is False

    def test_none_mode(self):
        group = ConditionGroup(
            conditions=[
                Condition("extension", "equals", "tmp"),
                Condition("extension", "equals", "temp"),
            ],
            mode=ConditionGroupMode.NONE,
        )

        assert group.evaluate({"extension": "pdf"}) is True
        assert group.evaluate({"extension": "tmp"}) is False

    def test_nested_groups(self):
        inner_group = ConditionGroup(
            conditions=[
                Condition("extension", "equals", "pdf"),
                Condition("extension", "equals", "doc"),
            ],
            mode=ConditionGroupMode.ANY,
        )

        outer_group = ConditionGroup(
            conditions=[
                inner_group,
                Condition("size", "greater_than", "1MB"),
            ],
            mode=ConditionGroupMode.ALL,
        )

        # PDF larger than 1MB
        assert outer_group.evaluate({"extension": "pdf", "size": 2 * 1024 * 1024}) is True

        # PDF smaller than 1MB
        assert outer_group.evaluate({"extension": "pdf", "size": 512 * 1024}) is False


class TestConditionSerialization:
    """Tests for condition serialization."""

    def test_to_dict(self):
        cond = Condition("extension", "equals", "pdf", case_sensitive=True)
        data = cond.to_dict()

        assert data["attribute"] == "extension"
        assert data["operator"] == "equals"
        assert data["value"] == "pdf"
        assert data["case_sensitive"] is True

    def test_from_dict(self):
        data = {
            "attribute": "name",
            "operator": "contains",
            "value": "report",
            "case_sensitive": False,
        }

        cond = Condition.from_dict(data)
        assert cond.evaluate({"name": "monthly_report"}) is True

    def test_condition_group_serialization(self):
        group = ConditionGroup(
            conditions=[
                Condition("extension", "equals", "pdf"),
                Condition("size", "greater_than", 1024),
            ],
            mode=ConditionGroupMode.ALL,
        )

        data = group.to_dict()
        restored = ConditionGroup.from_dict(data)

        assert len(restored.conditions) == 2
        assert restored.mode == ConditionGroupMode.ALL


class TestConvenienceFunctions:
    """Tests for condition convenience functions."""

    def test_name_equals(self):
        cond = name_equals("document")
        assert cond.evaluate({"name": "Document"}) is True

    def test_name_contains(self):
        cond = name_contains("report")
        assert cond.evaluate({"name": "monthly_report_2024"}) is True

    def test_extension_is(self):
        cond = extension_is(".pdf")  # With dot
        assert cond.evaluate({"extension": "pdf"}) is True

        cond = extension_is("pdf")  # Without dot
        assert cond.evaluate({"extension": "pdf"}) is True

    def test_extension_in(self):
        cond = extension_in(["pdf", "doc", "docx"])
        assert cond.evaluate({"extension": "pdf"}) is True
        assert cond.evaluate({"extension": "txt"}) is False

    def test_size_greater_than(self):
        cond = size_greater_than("5MB")
        assert cond.evaluate({"size": 10 * 1024 * 1024}) is True

    def test_size_less_than(self):
        cond = size_less_than("1MB")
        assert cond.evaluate({"size": 512 * 1024}) is True

    def test_modified_within(self):
        cond = modified_within("7 days")
        assert cond.evaluate({"date_modified": datetime.now() - timedelta(days=3)}) is True



"""
File information extraction.

This module provides functions for extracting various attributes from files,
including standard file system attributes and macOS-specific metadata.
"""

from __future__ import annotations

import mimetypes
import os
import subprocess
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional


def get_file_info(file_path: str) -> Dict[str, Any]:
    """
    Get comprehensive information about a file.

    Args:
        file_path: Path to the file.

    Returns:
        Dictionary containing file attributes.
    """
    path = Path(file_path)
    stat = path.stat()

    info = {
        # Basic attributes
        "path": str(path.absolute()),
        "name": path.stem,
        "extension": path.suffix.lstrip(".").lower(),
        "full_name": path.name,
        "parent_folder": path.parent.name,
        "parent_path": str(path.parent),

        # Size
        "size": stat.st_size,
        "size_bytes": stat.st_size,
        "size_human": _format_size(stat.st_size),

        # Dates
        "date_created": datetime.fromtimestamp(stat.st_birthtime) if hasattr(stat, "st_birthtime") else datetime.fromtimestamp(stat.st_ctime),
        "date_modified": datetime.fromtimestamp(stat.st_mtime),
        "date_accessed": datetime.fromtimestamp(stat.st_atime),

        # Type
        "is_file": path.is_file(),
        "is_directory": path.is_dir(),
        "is_symlink": path.is_symlink(),
        "is_hidden": path.name.startswith("."),
    }

    # MIME type
    mime_type, _ = mimetypes.guess_type(str(path))
    info["mime_type"] = mime_type
    info["file_type"] = mime_type

    # Determine kind based on extension/mime
    info["kind"] = _get_file_kind(path.suffix, mime_type)

    # macOS-specific attributes
    try:
        macos_info = _get_macos_metadata(file_path)
        info.update(macos_info)
    except Exception:
        pass

    return info


def _format_size(size_bytes: int) -> str:
    """Format size in bytes to human-readable string."""
    for unit in ["B", "KB", "MB", "GB", "TB"]:
        if size_bytes < 1024:
            return f"{size_bytes:.1f} {unit}"
        size_bytes /= 1024
    return f"{size_bytes:.1f} PB"


def _get_file_kind(extension: str, mime_type: Optional[str]) -> str:
    """Determine the kind of file based on extension and MIME type."""
    extension = extension.lower().lstrip(".")

    # Extension-based kinds
    kinds = {
        # Documents
        "pdf": "PDF Document",
        "doc": "Word Document",
        "docx": "Word Document",
        "xls": "Excel Spreadsheet",
        "xlsx": "Excel Spreadsheet",
        "ppt": "PowerPoint Presentation",
        "pptx": "PowerPoint Presentation",
        "txt": "Plain Text",
        "rtf": "Rich Text Document",
        "md": "Markdown Document",
        "pages": "Pages Document",
        "numbers": "Numbers Spreadsheet",
        "key": "Keynote Presentation",

        # Images
        "jpg": "JPEG Image",
        "jpeg": "JPEG Image",
        "png": "PNG Image",
        "gif": "GIF Image",
        "bmp": "BMP Image",
        "tiff": "TIFF Image",
        "tif": "TIFF Image",
        "webp": "WebP Image",
        "svg": "SVG Image",
        "ico": "Icon",
        "heic": "HEIC Image",
        "heif": "HEIF Image",
        "raw": "RAW Image",
        "psd": "Photoshop Document",
        "ai": "Illustrator Document",

        # Audio
        "mp3": "MP3 Audio",
        "wav": "WAV Audio",
        "aac": "AAC Audio",
        "flac": "FLAC Audio",
        "m4a": "M4A Audio",
        "ogg": "OGG Audio",
        "wma": "WMA Audio",
        "aiff": "AIFF Audio",

        # Video
        "mp4": "MP4 Video",
        "mov": "QuickTime Movie",
        "avi": "AVI Video",
        "mkv": "MKV Video",
        "wmv": "WMV Video",
        "flv": "FLV Video",
        "webm": "WebM Video",
        "m4v": "M4V Video",

        # Archives
        "zip": "ZIP Archive",
        "rar": "RAR Archive",
        "7z": "7-Zip Archive",
        "tar": "TAR Archive",
        "gz": "Gzip Archive",
        "bz2": "Bzip2 Archive",
        "dmg": "Disk Image",
        "iso": "ISO Image",

        # Code
        "py": "Python Script",
        "js": "JavaScript",
        "ts": "TypeScript",
        "html": "HTML Document",
        "css": "CSS Stylesheet",
        "json": "JSON File",
        "xml": "XML File",
        "yaml": "YAML File",
        "yml": "YAML File",
        "sh": "Shell Script",
        "java": "Java Source",
        "cpp": "C++ Source",
        "c": "C Source",
        "h": "Header File",
        "swift": "Swift Source",
        "rb": "Ruby Script",
        "go": "Go Source",
        "rs": "Rust Source",

        # Applications
        "app": "Application",
        "exe": "Windows Executable",
        "pkg": "Installer Package",

        # Other
        "torrent": "Torrent File",
        "ics": "Calendar Event",
        "vcf": "Contact Card",
    }

    if extension in kinds:
        return kinds[extension]

    # MIME-based fallback
    if mime_type:
        if mime_type.startswith("image/"):
            return "Image"
        elif mime_type.startswith("video/"):
            return "Video"
        elif mime_type.startswith("audio/"):
            return "Audio"
        elif mime_type.startswith("text/"):
            return "Text Document"
        elif mime_type.startswith("application/"):
            return "Application"

    return "Document"


def _get_macos_metadata(file_path: str) -> Dict[str, Any]:
    """
    Get macOS-specific metadata using mdls command.

    Args:
        file_path: Path to the file.

    Returns:
        Dictionary of macOS metadata.
    """
    info = {}

    try:
        # Get Spotlight metadata
        result = subprocess.run(
            ["mdls", "-plist", "-", file_path],
            capture_output=True,
            text=True,
            timeout=5,
        )

        if result.returncode == 0:
            import plistlib
            metadata = plistlib.loads(result.stdout.encode())

            # Extract useful attributes
            attr_mapping = {
                "kMDItemWhereFroms": "where_from",
                "kMDItemUserTags": "tags",
                "kMDItemFinderComment": "finder_comment",
                "kMDItemContentType": "uti",
                "kMDItemKind": "kind_macos",
                "kMDItemDateAdded": "date_added",
                "kMDItemDownloadedDate": "date_downloaded",
                "kMDItemPixelHeight": "pixel_height",
                "kMDItemPixelWidth": "pixel_width",
                "kMDItemDurationSeconds": "duration_seconds",
                "kMDItemTitle": "title",
                "kMDItemAuthors": "authors",
                "kMDItemCreator": "creator",
                "kMDItemDescription": "description",
            }

            for mdls_key, our_key in attr_mapping.items():
                if mdls_key in metadata and metadata[mdls_key] is not None:
                    value = metadata[mdls_key]
                    # Handle lists
                    if isinstance(value, list) and len(value) == 1:
                        value = value[0]
                    info[our_key] = value

    except (subprocess.TimeoutExpired, subprocess.SubprocessError):
        pass
    except Exception:
        pass

    # Get tags using xattr if not found via mdls
    if "tags" not in info:
        try:
            tags = _get_finder_tags(file_path)
            if tags:
                info["tags"] = tags
        except Exception:
            pass

    return info


def _get_finder_tags(file_path: str) -> List[str]:
    """
    Get Finder tags from a file using xattr.

    Args:
        file_path: Path to the file.

    Returns:
        List of tag names.
    """
    try:
        import plistlib

        result = subprocess.run(
            ["xattr", "-p", "com.apple.metadata:_kMDItemUserTags", file_path],
            capture_output=True,
            timeout=5,
        )

        if result.returncode == 0 and result.stdout:
            # Parse the binary plist
            tags_data = plistlib.loads(result.stdout)
            return [tag.split("\n")[0] for tag in tags_data if tag]
    except Exception:
        pass

    return []


def get_file_contents(file_path: str, max_bytes: int = 1024 * 1024) -> Optional[str]:
    """
    Read file contents (for text files).

    Args:
        file_path: Path to the file.
        max_bytes: Maximum bytes to read.

    Returns:
        File contents as string, or None if not readable.
    """
    try:
        with open(file_path, "r", encoding="utf-8", errors="ignore") as f:
            return f.read(max_bytes)
    except Exception:
        return None


def get_download_source(file_path: str) -> Optional[str]:
    """
    Get the URL from which a file was downloaded.

    Args:
        file_path: Path to the file.

    Returns:
        Download URL or None.
    """
    info = _get_macos_metadata(file_path)
    where_from = info.get("where_from")

    if isinstance(where_from, list):
        return where_from[0] if where_from else None
    return where_from


def is_complete_download(file_path: str) -> bool:
    """
    Check if a file is a complete download (not still being downloaded).

    Args:
        file_path: Path to the file.

    Returns:
        True if download is complete.
    """
    path = Path(file_path)

    # Check for common incomplete download extensions
    incomplete_extensions = {
        ".crdownload",  # Chrome
        ".part",  # Firefox, wget
        ".download",  # Safari
        ".tmp",
        ".partial",
    }

    if path.suffix.lower() in incomplete_extensions:
        return False

    # Check for .download companion file (Safari)
    if (path.parent / f"{path.name}.download").exists():
        return False

    return True



"""
Folder watching using macOS FSEvents.

This module provides file system monitoring for detecting changes in watched folders.
"""

from __future__ import annotations

import logging
import os
import threading
from pathlib import Path
from typing import Callable, Dict, List, Optional, Set
from datetime import datetime
import time

from watchdog.observers import Observer
from watchdog.events import (
    FileSystemEventHandler,
    FileCreatedEvent,
    FileModifiedEvent,
    FileMovedEvent,
    FileDeletedEvent,
    DirCreatedEvent,
    DirModifiedEvent,
    DirMovedEvent,
    DirDeletedEvent,
)

from sortmeout.utils.logger import get_logger

logger = get_logger(__name__)


class FileEventHandler(FileSystemEventHandler):
    """
    Handle file system events from watchdog.

    Filters events and calls the appropriate callback with normalized event types.
    """

    def __init__(
        self,
        callback: Callable[[str, str, str], None],
        folder_path: str,
        ignore_patterns: Optional[List[str]] = None,
        include_directories: bool = False,
        debounce_seconds: float = 0.5,
    ):
        """
        Initialize the event handler.

        Args:
            callback: Function to call on file events (event_type, file_path, folder_path).
            folder_path: Path to the watched folder.
            ignore_patterns: File patterns to ignore.
            include_directories: Whether to process directory events.
            debounce_seconds: Time to wait before processing rapid events.
        """
        super().__init__()
        self.callback = callback
        self.folder_path = folder_path
        self.ignore_patterns = ignore_patterns or [
            ".*",  # Hidden files
            "*.tmp",
            "*.temp",
            "*.part",
            "*.crdownload",
            ".DS_Store",
            "Thumbs.db",
        ]
        self.include_directories = include_directories
        self.debounce_seconds = debounce_seconds

        # Debouncing: track recent events
        self._recent_events: Dict[str, datetime] = {}
        self._lock = threading.Lock()

    def _should_ignore(self, path: str) -> bool:
        """Check if a path should be ignored."""
        import fnmatch

        name = os.path.basename(path)

        for pattern in self.ignore_patterns:
            if fnmatch.fnmatch(name, pattern):
                return True

        return False

    def _debounce(self, path: str) -> bool:
        """
        Check if event should be processed (debouncing).

        Returns True if event should be processed.
        """
        now = datetime.now()

        with self._lock:
            last_event = self._recent_events.get(path)

            if last_event:
                delta = (now - last_event).total_seconds()
                if delta < self.debounce_seconds:
                    return False

            self._recent_events[path] = now

            # Clean up old entries
            cutoff = now.timestamp() - 60  # Keep last minute
            self._recent_events = {
                p: t for p, t in self._recent_events.items()
                if t.timestamp() > cutoff
            }

        return True

    def _process_event(self, event_type: str, src_path: str, is_directory: bool) -> None:
        """Process a file system event."""
        if is_directory and not self.include_directories:
            return

        if self._should_ignore(src_path):
            return

        if not self._debounce(src_path):
            return

        logger.debug("Processing event: %s - %s", event_type, src_path)

        try:
            self.callback(event_type, src_path, self.folder_path)
        except Exception as e:
            logger.error("Error in callback for %s: %s", src_path, e)

    def on_created(self, event) -> None:
        """Handle file/directory created event."""
        # Wait a bit for file to be fully written
        time.sleep(0.1)
        self._process_event("created", event.src_path, event.is_directory)

    def on_modified(self, event) -> None:
        """Handle file/directory modified event."""
        self._process_event("modified", event.src_path, event.is_directory)

    def on_moved(self, event) -> None:
        """Handle file/directory moved event."""
        # Treat destination as created
        self._process_event("created", event.dest_path, event.is_directory)

    def on_deleted(self, event) -> None:
        """Handle file/directory deleted event."""
        self._process_event("deleted", event.src_path, event.is_directory)


class FolderWatcher:
    """
    Watch a single folder for file changes.

    Attributes:
        path: Path to the watched folder.
        recursive: Whether to watch subdirectories.
        enabled: Whether watching is active.
    """

    def __init__(
        self,
        path: str,
        callback: Callable[[str, str, str], None],
        recursive: bool = False,
        ignore_patterns: Optional[List[str]] = None,
    ):
        """
        Initialize a folder watcher.

        Args:
            path: Path to watch.
            callback: Function to call on file events.
            recursive: Watch subdirectories.
            ignore_patterns: Patterns to ignore.
        """
        self.path = os.path.expanduser(path)
        self.callback = callback
        self.recursive = recursive
        self.ignore_patterns = ignore_patterns
        self.enabled = True

        self._observer: Optional[Observer] = None
        self._handler: Optional[FileEventHandler] = None

        # Validate path
        if not os.path.isdir(self.path):
            raise ValueError(f"Path is not a directory: {self.path}")

    def start(self) -> None:
        """Start watching the folder."""
        if self._observer is not None:
            return

        self._handler = FileEventHandler(
            callback=self.callback,
            folder_path=self.path,
            ignore_patterns=self.ignore_patterns,
        )

        self._observer = Observer()
        self._observer.schedule(
            self._handler,
            self.path,
            recursive=self.recursive,
        )
        self._observer.start()

        logger.info("Started watching: %s (recursive=%s)", self.path, self.recursive)

    def stop(self) -> None:
        """Stop watching the folder."""
        if self._observer is None:
            return

        self._observer.stop()
        self._observer.join(timeout=5)
        self._observer = None
        self._handler = None

        logger.info("Stopped watching: %s", self.path)

    def is_running(self) -> bool:
        """Check if watcher is running."""
        return self._observer is not None and self._observer.is_alive()

    def __repr__(self) -> str:
        status = "running" if self.is_running() else "stopped"
        return f"FolderWatcher({self.path!r}, status={status})"


class WatcherManager:
    """
    Manage multiple folder watchers.

    Provides centralized control over all folder watchers and their lifecycle.
    """

    def __init__(self):
        """Initialize the watcher manager."""
        self._watchers: Dict[str, FolderWatcher] = {}
        self._running = False
        self._lock = threading.RLock()

    def add_watch(
        self,
        path: str,
        callback: Callable[[str, str, str], None],
        recursive: bool = False,
        ignore_patterns: Optional[List[str]] = None,
    ) -> bool:
        """
        Add a folder to watch.

        Args:
            path: Path to the folder.
            callback: Function to call on file events.
            recursive: Watch subdirectories.
            ignore_patterns: Patterns to ignore.

        Returns:
            True if watch was added successfully.
        """
        resolved_path = str(Path(os.path.expanduser(path)).resolve())

        with self._lock:
            if resolved_path in self._watchers:
                logger.warning("Already watching: %s", resolved_path)
                return False

            try:
                watcher = FolderWatcher(
                    path=resolved_path,
                    callback=callback,
                    recursive=recursive,
                    ignore_patterns=ignore_patterns,
                )

                if self._running:
                    watcher.start()

                self._watchers[resolved_path] = watcher
                logger.info("Added watch: %s", resolved_path)
                return True

            except Exception as e:
                logger.error("Failed to add watch for %s: %s", resolved_path, e)
                return False

    def remove_watch(self, path: str) -> bool:
        """
        Remove a folder watch.

        Args:
            path: Path to the folder.

        Returns:
            True if watch was removed successfully.
        """
        resolved_path = str(Path(os.path.expanduser(path)).resolve())

        with self._lock:
            if resolved_path not in self._watchers:
                return False

            watcher = self._watchers.pop(resolved_path)
            watcher.stop()

            logger.info("Removed watch: %s", resolved_path)
            return True

    def get_watched_folders(self) -> List[str]:
        """Get list of all watched folders."""
        with self._lock:
            return list(self._watchers.keys())

    def start(self) -> None:
        """Start all watchers."""
        with self._lock:
            if self._running:
                return

            self._running = True

            for watcher in self._watchers.values():
                if not watcher.is_running():
                    watcher.start()

            logger.info("Started %d watchers", len(self._watchers))

    def stop(self) -> None:
        """Stop all watchers."""
        with self._lock:
            if not self._running:
                return

            self._running = False

            for watcher in self._watchers.values():
                watcher.stop()

            logger.info("Stopped all watchers")

    def is_running(self) -> bool:
        """Check if manager is running."""
        return self._running

    def get_status(self) -> Dict[str, Dict]:
        """Get status of all watchers."""
        with self._lock:
            return {
                path: {
                    "running": watcher.is_running(),
                    "recursive": watcher.recursive,
                    "enabled": watcher.enabled,
                }
                for path, watcher in self._watchers.items()
            }

    def __len__(self) -> int:
        """Number of watched folders."""
        return len(self._watchers)

    def __repr__(self) -> str:
        status = "running" if self._running else "stopped"
        return f"WatcherManager({len(self._watchers)} watchers, status={status})"


class FSEventsWatcher:
    """
    Native macOS FSEvents watcher using PyObjC.

    This provides more efficient file system monitoring on macOS
    compared to the generic watchdog implementation.
    """

    def __init__(
        self,
        path: str,
        callback: Callable[[str, str, str], None],
        latency: float = 0.5,
    ):
        """
        Initialize FSEvents watcher.

        Args:
            path: Path to watch.
            callback: Callback function.
            latency: Latency in seconds before reporting events.
        """
        self.path = os.path.expanduser(path)
        self.callback = callback
        self.latency = latency
        self._stream = None
        self._running = False

    def start(self) -> None:
        """Start the FSEvents stream."""
        try:
            from FSEvents import (
                FSEventStreamCreate,
                FSEventStreamScheduleWithRunLoop,
                FSEventStreamStart,
                FSEventStreamStop,
                FSEventStreamInvalidate,
                FSEventStreamRelease,
                CFRunLoopGetCurrent,
                kCFRunLoopDefaultMode,
                kFSEventStreamEventFlagNone,
            )

            def fsevents_callback(stream, info, num_events, paths, flags, ids):
                for i in range(num_events):
                    path = paths[i].decode() if isinstance(paths[i], bytes) else paths[i]
                    self.callback("modified", path, self.path)

            self._stream = FSEventStreamCreate(
                None,
                fsevents_callback,
                None,
                [self.path],
                -1,  # kFSEventStreamEventIdSinceNow
                self.latency,
                kFSEventStreamEventFlagNone,
            )

            FSEventStreamScheduleWithRunLoop(
                self._stream,
                CFRunLoopGetCurrent(),
                kCFRunLoopDefaultMode,
            )

            FSEventStreamStart(self._stream)
            self._running = True

            logger.info("Started FSEvents watcher: %s", self.path)

        except ImportError:
            logger.warning("FSEvents not available, falling back to watchdog")
            raise

    def stop(self) -> None:
        """Stop the FSEvents stream."""
        if self._stream:
            try:
                from FSEvents import (
                    FSEventStreamStop,
                    FSEventStreamInvalidate,
                    FSEventStreamRelease,
                )

                FSEventStreamStop(self._stream)
                FSEventStreamInvalidate(self._stream)
                FSEventStreamRelease(self._stream)

            except ImportError:
                pass

            self._stream = None
            self._running = False

            logger.info("Stopped FSEvents watcher: %s", self.path)

    def is_running(self) -> bool:
        """Check if watcher is running."""
        return self._running



"""
Tests for the Rule class.
"""

import pytest
from datetime import datetime

from sortmeout.core.rule import Rule, MatchMode
from sortmeout.core.condition import Condition, ConditionGroup
from sortmeout.core.action import Action, ActionType


class TestRuleCreation:
    """Tests for rule creation."""

    def test_create_rule_with_name(self):
        rule = Rule(name="My Rule")
        assert rule.name == "My Rule"
        assert rule.enabled is True

    def test_create_rule_with_defaults(self):
        rule = Rule(name="Test")
        assert rule.match_mode == MatchMode.ALL
        assert rule.enabled is True
        assert rule.conditions == []
        assert rule.actions == []
        assert rule.priority == 0

    def test_rule_has_unique_id(self):
        rule1 = Rule(name="Rule 1")
        rule2 = Rule(name="Rule 2")
        assert rule1.id != rule2.id


class TestMatchMode:
    """Tests for match modes."""

    def test_all_match_mode(self):
        rule = Rule(
            name="Test",
            match_mode=MatchMode.ALL,
            conditions=[
                Condition("name", "contains", "test"),
                Condition("extension", "equals", "txt"),
            ]
        )

        # Both match
        file_info = {"name": "test_file", "extension": "txt"}
        assert rule.matches(file_info)

        # Only one matches
        file_info = {"name": "test_file", "extension": "pdf"}
        assert not rule.matches(file_info)

    def test_any_match_mode(self):
        rule = Rule(
            name="Test",
            match_mode=MatchMode.ANY,
            conditions=[
                Condition("name", "contains", "test"),
                Condition("extension", "equals", "txt"),
            ]
        )

        # One matches
        file_info = {"name": "test_file", "extension": "pdf"}
        assert rule.matches(file_info)

        # Neither matches
        file_info = {"name": "file", "extension": "pdf"}
        assert not rule.matches(file_info)

    def test_none_match_mode(self):
        rule = Rule(
            name="Test",
            match_mode=MatchMode.NONE,
            conditions=[
                Condition("name", "contains", "test"),
                Condition("extension", "equals", "txt"),
            ]
        )

        # Neither matches
        file_info = {"name": "file", "extension": "pdf"}
        assert rule.matches(file_info)

        # One matches
        file_info = {"name": "test_file", "extension": "pdf"}
        assert not rule.matches(file_info)


class TestRuleConditions:
    """Tests for rule conditions management."""

    def test_add_condition(self):
        rule = Rule(name="Test")
        condition = Condition("name", "contains", "test")
        rule.add_condition(condition)

        assert len(rule.conditions) == 1
        assert rule.conditions[0] == condition

    def test_add_multiple_conditions(self):
        rule = Rule(name="Test")
        rule.add_condition(Condition("name", "contains", "test"))
        rule.add_condition(Condition("extension", "equals", "txt"))

        assert len(rule.conditions) == 2

    def test_remove_condition(self):
        condition = Condition("name", "contains", "test")
        rule = Rule(name="Test", conditions=[condition])

        rule.remove_condition(condition.id)

        assert len(rule.conditions) == 0

    def test_clear_conditions(self):
        rule = Rule(
            name="Test",
            conditions=[
                Condition("name", "contains", "test"),
                Condition("extension", "equals", "txt"),
            ]
        )

        rule.clear_conditions()

        assert len(rule.conditions) == 0


class TestRuleActions:
    """Tests for rule actions management."""

    def test_add_action(self):
        rule = Rule(name="Test")
        action = Action("move", destination="~/Documents")
        rule.add_action(action)

        assert len(rule.actions) == 1
        assert rule.actions[0] == action

    def test_add_multiple_actions(self):
        rule = Rule(name="Test")
        rule.add_action(Action("move", destination="~/Documents"))
        rule.add_action(Action("add_tags", tags=["processed"]))

        assert len(rule.actions) == 2

    def test_actions_preserve_order(self):
        rule = Rule(name="Test")
        action1 = Action("add_tags", tags=["step1"])
        action2 = Action("move", destination="~/Documents")
        action3 = Action("notify", message="Done")

        rule.add_action(action1)
        rule.add_action(action2)
        rule.add_action(action3)

        assert rule.actions[0] == action1
        assert rule.actions[1] == action2
        assert rule.actions[2] == action3

    def test_remove_action(self):
        action = Action("move", destination="~/Documents")
        rule = Rule(name="Test", actions=[action])

        rule.remove_action(action.id)

        assert len(rule.actions) == 0

    def test_clear_actions(self):
        rule = Rule(
            name="Test",
            actions=[
                Action("move", destination="~/Documents"),
                Action("notify", message="Done"),
            ]
        )

        rule.clear_actions()

        assert len(rule.actions) == 0


class TestRuleMatching:
    """Tests for rule matching."""

    def test_matches_file_info(self):
        rule = Rule(
            name="Text Files",
            conditions=[
                Condition("extension", "equals", "txt"),
            ]
        )

        assert rule.matches({"extension": "txt"})
        assert not rule.matches({"extension": "pdf"})

    def test_matches_with_condition_group(self):
        group = ConditionGroup(
            match_mode="any",
            conditions=[
                Condition("extension", "equals", "txt"),
                Condition("extension", "equals", "md"),
            ]
        )

        rule = Rule(
            name="Text Files",
            conditions=[group]
        )

        assert rule.matches({"extension": "txt"})
        assert rule.matches({"extension": "md"})
        assert not rule.matches({"extension": "pdf"})

    def test_disabled_rule_never_matches(self):
        rule = Rule(
            name="Test",
            enabled=False,
            conditions=[Condition("extension", "equals", "txt")]
        )

        assert not rule.matches({"extension": "txt"})

    def test_rule_without_conditions_matches_all(self):
        rule = Rule(name="Catch All")

        assert rule.matches({"extension": "txt"})
        assert rule.matches({"extension": "pdf"})
        assert rule.matches({})


class TestRulePriority:
    """Tests for rule priority."""

    def test_priority_sorting(self):
        rule1 = Rule(name="Low", priority=1)
        rule2 = Rule(name="High", priority=10)
        rule3 = Rule(name="Medium", priority=5)

        rules = [rule1, rule2, rule3]
        sorted_rules = sorted(rules, key=lambda r: r.priority, reverse=True)

        assert sorted_rules[0].name == "High"
        assert sorted_rules[1].name == "Medium"
        assert sorted_rules[2].name == "Low"


class TestRuleSerialization:
    """Tests for rule serialization."""

    def test_to_dict(self):
        rule = Rule(
            name="Test Rule",
            match_mode=MatchMode.ALL,
            conditions=[Condition("extension", "equals", "txt")],
            actions=[Action("move", destination="~/Documents")],
            priority=5,
            enabled=True,
        )

        data = rule.to_dict()

        assert data["name"] == "Test Rule"
        assert data["match_mode"] == "all"
        assert len(data["conditions"]) == 1
        assert len(data["actions"]) == 1
        assert data["priority"] == 5
        assert data["enabled"] is True

    def test_from_dict(self):
        data = {
            "name": "Test Rule",
            "match_mode": "any",
            "conditions": [
                {
                    "attribute": "extension",
                    "operator": "equals",
                    "value": "txt",
                }
            ],
            "actions": [
                {
                    "action_type": "move",
                    "params": {"destination": "~/Documents"},
                }
            ],
            "priority": 3,
            "enabled": False,
        }

        rule = Rule.from_dict(data)

        assert rule.name == "Test Rule"
        assert rule.match_mode == MatchMode.ANY
        assert len(rule.conditions) == 1
        assert len(rule.actions) == 1
        assert rule.priority == 3
        assert rule.enabled is False

    def test_duplicate(self):
        original = Rule(
            name="Original",
            conditions=[Condition("extension", "equals", "txt")],
            actions=[Action("move", destination="~/Documents")],
        )

        duplicate = original.duplicate()

        assert duplicate.name == "Original (Copy)"
        assert duplicate.id != original.id
        assert len(duplicate.conditions) == len(original.conditions)
        assert len(duplicate.actions) == len(original.actions)


class TestRuleStats:
    """Tests for rule statistics."""

    def test_initial_stats(self):
        rule = Rule(name="Test")

        assert rule.run_count == 0
        assert rule.last_run is None

    def test_record_run(self):
        rule = Rule(name="Test")

        rule.record_run()

        assert rule.run_count == 1
        assert rule.last_run is not None
        assert isinstance(rule.last_run, datetime)

    def test_multiple_runs(self):
        rule = Rule(name="Test")

        rule.record_run()
        first_run = rule.last_run

        rule.record_run()
        second_run = rule.last_run

        assert rule.run_count == 2
        assert second_run >= first_run


class TestRuleDescription:
    """Tests for rule description generation."""

    def test_simple_description(self):
        rule = Rule(
            name="Test",
            conditions=[Condition("extension", "equals", "txt")],
            actions=[Action("move", destination="~/Documents")],
        )

        desc = rule.describe()

        assert "extension" in desc.lower()
        assert "move" in desc.lower()

    def test_empty_rule_description(self):
        rule = Rule(name="Empty")

        desc = rule.describe()

        assert "no conditions" in desc.lower()


class TestRuleValidation:
    """Tests for rule validation."""

    def test_validate_complete_rule(self):
        rule = Rule(
            name="Valid Rule",
            conditions=[Condition("extension", "equals", "txt")],
            actions=[Action("move", destination="~/Documents")],
        )

        errors = rule.validate()

        assert len(errors) == 0

    def test_validate_missing_name(self):
        rule = Rule(name="")

        errors = rule.validate()

        assert any("name" in e.lower() for e in errors)

    def test_validate_no_actions(self):
        rule = Rule(
            name="No Actions",
            conditions=[Condition("extension", "equals", "txt")],
        )

        errors = rule.validate()

        assert any("action" in e.lower() for e in errors)



"""
macOS Spotlight integration.

Provides functions for searching and retrieving Spotlight metadata.
"""

from __future__ import annotations

import plistlib
import subprocess
from datetime import datetime
from typing import Any, Dict, List, Optional

from sortmeout.utils.logger import get_logger

logger = get_logger(__name__)


def get_metadata(file_path: str) -> Dict[str, Any]:
    """
    Get all Spotlight metadata for a file.

    Args:
        file_path: Path to the file.

    Returns:
        Dictionary of metadata attributes.
    """
    try:
        result = subprocess.run(
            ["mdls", "-plist", "-", file_path],
            capture_output=True,
            timeout=10,
        )

        if result.returncode != 0:
            return {}

        metadata = plistlib.loads(result.stdout)
        return metadata

    except Exception as e:
        logger.error("Failed to get metadata for %s: %s", file_path, e)
        return {}


def get_attribute(file_path: str, attribute: str) -> Optional[Any]:
    """
    Get a specific Spotlight attribute for a file.

    Args:
        file_path: Path to the file.
        attribute: Spotlight attribute name (e.g., "kMDItemKind").

    Returns:
        Attribute value or None.
    """
    try:
        result = subprocess.run(
            ["mdls", "-name", attribute, "-raw", file_path],
            capture_output=True,
            text=True,
            timeout=10,
        )

        if result.returncode != 0:
            return None

        value = result.stdout.strip()
        if value == "(null)":
            return None

        return value

    except Exception as e:
        logger.debug("Failed to get attribute %s: %s", attribute, e)
        return None


def search_spotlight(
    query: str,
    folder: Optional[str] = None,
    limit: int = 100,
    attributes: Optional[List[str]] = None,
) -> List[Dict[str, Any]]:
    """
    Search for files using Spotlight.

    Args:
        query: Spotlight query string (e.g., "kMDItemKind == 'PDF Document'").
        folder: Optional folder to search in.
        limit: Maximum number of results.
        attributes: Specific attributes to retrieve.

    Returns:
        List of matching files with their metadata.
    """
    try:
        cmd = ["mdfind"]

        if folder:
            cmd.extend(["-onlyin", folder])

        cmd.append(query)

        result = subprocess.run(
            cmd,
            capture_output=True,
            text=True,
            timeout=30,
        )

        if result.returncode != 0:
            return []

        files = result.stdout.strip().split("\n")[:limit]

        if not attributes:
            return [{"path": f} for f in files if f]

        # Get attributes for each file
        results = []
        for file_path in files:
            if not file_path:
                continue

            metadata = {"path": file_path}
            for attr in attributes:
                value = get_attribute(file_path, attr)
                if value is not None:
                    metadata[attr] = value

            results.append(metadata)

        return results

    except Exception as e:
        logger.error("Spotlight search failed: %s", e)
        return []


def find_by_kind(kind: str, folder: Optional[str] = None, limit: int = 100) -> List[str]:
    """
    Find files by their kind (e.g., "PDF Document", "JPEG image").

    Args:
        kind: File kind to search for.
        folder: Optional folder to search in.
        limit: Maximum number of results.

    Returns:
        List of file paths.
    """
    query = f'kMDItemKind == "{kind}"'
    results = search_spotlight(query, folder, limit)
    return [r["path"] for r in results]


def find_by_extension(extension: str, folder: Optional[str] = None, limit: int = 100) -> List[str]:
    """
    Find files by extension.

    Args:
        extension: File extension (without dot).
        folder: Optional folder to search in.
        limit: Maximum number of results.

    Returns:
        List of file paths.
    """
    extension = extension.lstrip(".")
    query = f'kMDItemFSName == "*.{extension}"'
    results = search_spotlight(query, folder, limit)
    return [r["path"] for r in results]


def find_by_tag(tag: str, folder: Optional[str] = None, limit: int = 100) -> List[str]:
    """
    Find files by Finder tag.

    Args:
        tag: Tag name to search for.
        folder: Optional folder to search in.
        limit: Maximum number of results.

    Returns:
        List of file paths.
    """
    query = f'kMDItemUserTags == "{tag}"'
    results = search_spotlight(query, folder, limit)
    return [r["path"] for r in results]


def find_by_content(text: str, folder: Optional[str] = None, limit: int = 100) -> List[str]:
    """
    Find files containing specific text.

    Args:
        text: Text to search for in file contents.
        folder: Optional folder to search in.
        limit: Maximum number of results.

    Returns:
        List of file paths.
    """
    query = f'kMDItemTextContent == "{text}"c'  # c for case-insensitive
    results = search_spotlight(query, folder, limit)
    return [r["path"] for r in results]


def find_modified_after(date: datetime, folder: Optional[str] = None, limit: int = 100) -> List[str]:
    """
    Find files modified after a specific date.

    Args:
        date: Cutoff date.
        folder: Optional folder to search in.
        limit: Maximum number of results.

    Returns:
        List of file paths.
    """
    date_str = date.strftime("%Y-%m-%d")
    query = f'kMDItemFSContentChangeDate >= $time.iso({date_str})'
    results = search_spotlight(query, folder, limit)
    return [r["path"] for r in results]


def find_created_after(date: datetime, folder: Optional[str] = None, limit: int = 100) -> List[str]:
    """
    Find files created after a specific date.

    Args:
        date: Cutoff date.
        folder: Optional folder to search in.
        limit: Maximum number of results.

    Returns:
        List of file paths.
    """
    date_str = date.strftime("%Y-%m-%d")
    query = f'kMDItemFSCreationDate >= $time.iso({date_str})'
    results = search_spotlight(query, folder, limit)
    return [r["path"] for r in results]


def find_by_author(author: str, folder: Optional[str] = None, limit: int = 100) -> List[str]:
    """
    Find files by author.

    Args:
        author: Author name to search for.
        folder: Optional folder to search in.
        limit: Maximum number of results.

    Returns:
        List of file paths.
    """
    query = f'kMDItemAuthors == "{author}"c'
    results = search_spotlight(query, folder, limit)
    return [r["path"] for r in results]


def get_download_source(file_path: str) -> Optional[str]:
    """
    Get the URL from which a file was downloaded.

    Args:
        file_path: Path to the file.

    Returns:
        Download URL or None.
    """
    metadata = get_metadata(file_path)
    where_froms = metadata.get("kMDItemWhereFroms", [])

    if where_froms:
        return where_froms[0] if isinstance(where_froms, list) else where_froms

    return None


def get_uti(file_path: str) -> Optional[str]:
    """
    Get the Uniform Type Identifier for a file.

    Args:
        file_path: Path to the file.

    Returns:
        UTI string or None.
    """
    return get_attribute(file_path, "kMDItemContentType")


def file_matches_uti(file_path: str, uti: str) -> bool:
    """
    Check if a file conforms to a UTI.

    Args:
        file_path: Path to the file.
        uti: UTI to check against.

    Returns:
        True if file conforms to UTI.
    """
    file_uti = get_uti(file_path)
    if not file_uti:
        return False

    # Check if file UTI equals or conforms to target UTI
    if file_uti == uti:
        return True

    # Use mdls to check conformance
    type_tree = get_attribute(file_path, "kMDItemContentTypeTree")
    if type_tree:
        if isinstance(type_tree, str):
            type_tree = [type_tree]
        return uti in type_tree

    return False


# Common UTI constants
UTI_PDF = "com.adobe.pdf"
UTI_IMAGE = "public.image"
UTI_MOVIE = "public.movie"
UTI_AUDIO = "public.audio"
UTI_TEXT = "public.plain-text"
UTI_RTF = "public.rtf"
UTI_HTML = "public.html"
UTI_XML = "public.xml"
UTI_JSON = "public.json"
UTI_ZIP = "com.pkware.zip-archive"
UTI_FOLDER = "public.folder"
UTI_APPLICATION = "com.apple.application-bundle"



"""
Tests for the Action class.
"""

import os
import shutil
import tempfile
import pytest
from pathlib import Path

from sortmeout.core.action import (
    Action,
    ActionType,
    ActionResult,
    move_to,
    copy_to,
    rename,
    delete,
    trash,
    archive,
    add_tags,
    notify,
)


@pytest.fixture
def temp_dir():
    """Create a temporary directory for tests."""
    dir_path = tempfile.mkdtemp()
    yield dir_path
    # Cleanup
    if os.path.exists(dir_path):
        shutil.rmtree(dir_path)


@pytest.fixture
def test_file(temp_dir):
    """Create a test file."""
    file_path = os.path.join(temp_dir, "test_file.txt")
    with open(file_path, "w") as f:
        f.write("Test content")
    return file_path


@pytest.fixture
def file_info(test_file):
    """Generate basic file info dict."""
    path = Path(test_file)
    return {
        "path": str(path),
        "name": path.stem,
        "extension": path.suffix.lstrip("."),
        "full_name": path.name,
        "parent_folder": path.parent.name,
    }


class TestActionCreation:
    """Tests for action creation."""

    def test_create_with_type_string(self):
        action = Action("move", destination="~/Documents")
        assert action.action_type == ActionType.MOVE
        assert action.params["destination"] == "~/Documents"

    def test_create_with_type_enum(self):
        action = Action(ActionType.COPY, destination="~/Documents")
        assert action.action_type == ActionType.COPY

    def test_action_has_unique_id(self):
        action1 = Action("move", destination="~/Documents")
        action2 = Action("move", destination="~/Documents")
        assert action1.id != action2.id


class TestMoveAction:
    """Tests for move action."""

    def test_move_to_directory(self, temp_dir, test_file, file_info):
        dest_dir = os.path.join(temp_dir, "destination")
        os.makedirs(dest_dir)

        action = Action("move", destination=dest_dir)
        result = action.execute(test_file, file_info)

        assert result.success
        assert not os.path.exists(test_file)
        assert os.path.exists(os.path.join(dest_dir, "test_file.txt"))

    def test_move_creates_destination(self, temp_dir, test_file, file_info):
        dest_dir = os.path.join(temp_dir, "new_folder", "subfolder")

        action = Action("move", destination=dest_dir)
        result = action.execute(test_file, file_info)

        assert result.success
        assert os.path.exists(os.path.join(dest_dir, "test_file.txt"))

    def test_move_handles_conflict_rename(self, temp_dir, test_file, file_info):
        dest_dir = os.path.join(temp_dir, "destination")
        os.makedirs(dest_dir)

        # Create existing file
        existing = os.path.join(dest_dir, "test_file.txt")
        with open(existing, "w") as f:
            f.write("Existing")

        action = Action("move", destination=dest_dir, if_exists="rename")
        result = action.execute(test_file, file_info)

        assert result.success
        assert os.path.exists(existing)  # Original still exists
        assert "test_file (1).txt" in result.destination_path

    def test_move_preview_mode(self, test_file, file_info):
        action = Action("move", destination="~/Documents")
        result = action.execute(test_file, file_info, preview=True)

        assert result.success
        assert os.path.exists(test_file)  # File not actually moved
        assert "PREVIEW" in result.message


class TestCopyAction:
    """Tests for copy action."""

    def test_copy_to_directory(self, temp_dir, test_file, file_info):
        dest_dir = os.path.join(temp_dir, "destination")
        os.makedirs(dest_dir)

        action = Action("copy", destination=dest_dir)
        result = action.execute(test_file, file_info)

        assert result.success
        assert os.path.exists(test_file)  # Original still exists
        assert os.path.exists(os.path.join(dest_dir, "test_file.txt"))

    def test_copy_preserves_metadata(self, temp_dir, test_file, file_info):
        dest_dir = os.path.join(temp_dir, "destination")
        os.makedirs(dest_dir)

        action = Action("copy", destination=dest_dir)
        result = action.execute(test_file, file_info)

        original_stat = os.stat(test_file)
        copy_stat = os.stat(result.destination_path)

        assert original_stat.st_mode == copy_stat.st_mode


class TestRenameAction:
    """Tests for rename action."""

    def test_rename_file(self, temp_dir, test_file, file_info):
        action = Action("rename", new_name="renamed_file.txt")
        result = action.execute(test_file, file_info)

        assert result.success
        assert not os.path.exists(test_file)
        assert os.path.exists(os.path.join(temp_dir, "renamed_file.txt"))

    def test_rename_with_variables(self, temp_dir, test_file, file_info):
        action = Action("rename", new_name="{name}_processed.{extension}")
        result = action.execute(test_file, file_info)

        assert result.success
        assert "test_file_processed.txt" in result.destination_path

    def test_rename_handles_conflict(self, temp_dir, test_file, file_info):
        # Create a file with the target name
        target = os.path.join(temp_dir, "target.txt")
        with open(target, "w") as f:
            f.write("Existing")

        action = Action("rename", new_name="target.txt", if_exists="rename")
        result = action.execute(test_file, file_info)

        assert result.success
        assert "target (1).txt" in result.destination_path


class TestDeleteAction:
    """Tests for delete action."""

    def test_delete_file(self, test_file, file_info):
        action = Action("delete", force=True)
        result = action.execute(test_file, file_info)

        assert result.success
        assert not os.path.exists(test_file)

    def test_delete_directory(self, temp_dir, file_info):
        dir_to_delete = os.path.join(temp_dir, "to_delete")
        os.makedirs(dir_to_delete)

        # Add some files
        with open(os.path.join(dir_to_delete, "file.txt"), "w") as f:
            f.write("Content")

        action = Action("delete", force=True)
        result = action.execute(dir_to_delete, {"path": dir_to_delete})

        assert result.success
        assert not os.path.exists(dir_to_delete)


class TestArchiveAction:
    """Tests for archive action."""

    def test_archive_zip(self, temp_dir, test_file, file_info):
        action = Action("archive", format="zip")
        result = action.execute(test_file, file_info)

        assert result.success
        assert os.path.exists(test_file)  # Original still exists
        assert result.destination_path.endswith(".zip")

    def test_archive_delete_original(self, temp_dir, test_file, file_info):
        action = Action("archive", format="zip", delete_original=True)
        result = action.execute(test_file, file_info)

        assert result.success
        assert not os.path.exists(test_file)


class TestVariableExpansion:
    """Tests for variable expansion in action parameters."""

    def test_expand_name(self, test_file, file_info):
        action = Action("rename", new_name="prefix_{name}.txt")
        result = action.execute(test_file, file_info, preview=True)

        assert "prefix_test_file.txt" in str(result.metadata.get("params", {}).get("new_name", ""))

    def test_expand_extension(self, test_file, file_info):
        action = Action("rename", new_name="file.{extension}")
        result = action.execute(test_file, file_info, preview=True)

        params = result.metadata.get("params", {})
        assert "file.txt" in str(params.get("new_name", ""))

    def test_expand_date(self, test_file, file_info):
        from datetime import datetime

        action = Action("rename", new_name="{name}_{date}.txt")
        result = action.execute(test_file, file_info, preview=True)

        today = datetime.now().strftime("%Y-%m-%d")
        params = result.metadata.get("params", {})
        assert today in str(params.get("new_name", ""))


class TestActionSerialization:
    """Tests for action serialization."""

    def test_to_dict(self):
        action = Action(
            "move",
            destination="~/Documents",
            enabled=True,
            stop_on_error=False,
        )

        data = action.to_dict()

        assert data["action_type"] == "move"
        assert data["params"]["destination"] == "~/Documents"
        assert data["enabled"] is True
        assert data["stop_on_error"] is False

    def test_from_dict(self):
        data = {
            "action_type": "copy",
            "params": {"destination": "~/Backup"},
            "enabled": True,
        }

        action = Action.from_dict(data)

        assert action.action_type == ActionType.COPY
        assert action.params["destination"] == "~/Backup"

    def test_duplicate(self):
        original = Action("move", destination="~/Documents")
        duplicate = original.duplicate()

        assert duplicate.action_type == original.action_type
        assert duplicate.params == original.params
        assert duplicate.id != original.id


class TestConvenienceFunctions:
    """Tests for action convenience functions."""

    def test_move_to(self):
        action = move_to("~/Documents")
        assert action.action_type == ActionType.MOVE
        assert action.params["destination"] == "~/Documents"

    def test_copy_to(self):
        action = copy_to("~/Backup")
        assert action.action_type == ActionType.COPY

    def test_rename(self):
        action = rename("new_name.txt")
        assert action.action_type == ActionType.RENAME
        assert action.params["new_name"] == "new_name.txt"

    def test_delete(self):
        action = delete(force=True)
        assert action.action_type == ActionType.DELETE
        assert action.params["force"] is True

    def test_trash(self):
        action = trash()
        assert action.action_type == ActionType.TRASH

    def test_archive(self):
        action = archive(format="zip", delete_original=True)
        assert action.action_type == ActionType.ARCHIVE
        assert action.params["format"] == "zip"
        assert action.params["delete_original"] is True

    def test_add_tags(self):
        action = add_tags("work", "important")
        assert action.action_type == ActionType.ADD_TAGS
        assert action.params["tags"] == ["work", "important"]


class TestActionResult:
    """Tests for ActionResult."""

    def test_success_result(self):
        result = ActionResult(
            success=True,
            action_type=ActionType.MOVE,
            source_path="/source/file.txt",
            destination_path="/dest/file.txt",
            message="Moved successfully",
        )

        assert result.success
        assert "✓" in str(result)

    def test_failure_result(self):
        result = ActionResult(
            success=False,
            action_type=ActionType.MOVE,
            source_path="/source/file.txt",
            error="Permission denied",
        )

        assert not result.success
        assert "✗" in str(result)



"""
Application settings.

User-configurable settings for SortMeOut behavior.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional


@dataclass
class TrashSettings:
    """Settings for trash management."""
    enabled: bool = False
    max_age_days: int = 30
    max_size_gb: float = 10.0
    app_sweep_enabled: bool = True
    app_sweep_prompt: bool = True  # Ask before cleaning app files

    def to_dict(self) -> Dict[str, Any]:
        return {
            "enabled": self.enabled,
            "max_age_days": self.max_age_days,
            "max_size_gb": self.max_size_gb,
            "app_sweep_enabled": self.app_sweep_enabled,
            "app_sweep_prompt": self.app_sweep_prompt,
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "TrashSettings":
        return cls(
            enabled=data.get("enabled", False),
            max_age_days=data.get("max_age_days", 30),
            max_size_gb=data.get("max_size_gb", 10.0),
            app_sweep_enabled=data.get("app_sweep_enabled", True),
            app_sweep_prompt=data.get("app_sweep_prompt", True),
        )


@dataclass
class NotificationSettings:
    """Settings for notifications."""
    enabled: bool = True
    show_rule_matches: bool = False
    show_errors: bool = True
    show_summary: bool = True
    summary_interval_minutes: int = 60
    sound_enabled: bool = True

    def to_dict(self) -> Dict[str, Any]:
        return {
            "enabled": self.enabled,
            "show_rule_matches": self.show_rule_matches,
            "show_errors": self.show_errors,
            "show_summary": self.show_summary,
            "summary_interval_minutes": self.summary_interval_minutes,
            "sound_enabled": self.sound_enabled,
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "NotificationSettings":
        return cls(
            enabled=data.get("enabled", True),
            show_rule_matches=data.get("show_rule_matches", False),
            show_errors=data.get("show_errors", True),
            show_summary=data.get("show_summary", True),
            summary_interval_minutes=data.get("summary_interval_minutes", 60),
            sound_enabled=data.get("sound_enabled", True),
        )


@dataclass
class LoggingSettings:
    """Settings for logging."""
    level: str = "INFO"
    file_logging: bool = True
    max_log_size_mb: int = 10
    backup_count: int = 5
    action_logging: bool = True

    def to_dict(self) -> Dict[str, Any]:
        return {
            "level": self.level,
            "file_logging": self.file_logging,
            "max_log_size_mb": self.max_log_size_mb,
            "backup_count": self.backup_count,
            "action_logging": self.action_logging,
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "LoggingSettings":
        return cls(
            level=data.get("level", "INFO"),
            file_logging=data.get("file_logging", True),
            max_log_size_mb=data.get("max_log_size_mb", 10),
            backup_count=data.get("backup_count", 5),
            action_logging=data.get("action_logging", True),
        )


@dataclass
class WatcherSettings:
    """Settings for file watching."""
    latency_seconds: float = 0.5
    debounce_seconds: float = 0.5
    ignore_hidden_files: bool = True
    ignore_system_files: bool = True
    custom_ignore_patterns: List[str] = field(default_factory=list)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "latency_seconds": self.latency_seconds,
            "debounce_seconds": self.debounce_seconds,
            "ignore_hidden_files": self.ignore_hidden_files,
            "ignore_system_files": self.ignore_system_files,
            "custom_ignore_patterns": self.custom_ignore_patterns,
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "WatcherSettings":
        return cls(
            latency_seconds=data.get("latency_seconds", 0.5),
            debounce_seconds=data.get("debounce_seconds", 0.5),
            ignore_hidden_files=data.get("ignore_hidden_files", True),
            ignore_system_files=data.get("ignore_system_files", True),
            custom_ignore_patterns=data.get("custom_ignore_patterns", []),
        )


@dataclass
class Settings:
    """
    Main application settings.

    Contains all user-configurable settings for SortMeOut.
    """
    # General
    start_at_login: bool = False
    show_menu_bar_icon: bool = True
    preview_mode: bool = False
    confirm_destructive_actions: bool = True

    # Subsystems
    trash: TrashSettings = field(default_factory=TrashSettings)
    notifications: NotificationSettings = field(default_factory=NotificationSettings)
    logging: LoggingSettings = field(default_factory=LoggingSettings)
    watcher: WatcherSettings = field(default_factory=WatcherSettings)

    # UI preferences
    theme: str = "system"  # system, light, dark
    language: str = "en"
    show_preview_on_hover: bool = True

    # Advanced
    max_concurrent_actions: int = 5
    action_timeout_seconds: int = 300
    retry_failed_actions: bool = False
    retry_count: int = 3

    def to_dict(self) -> Dict[str, Any]:
        """Convert settings to dictionary."""
        return {
            "start_at_login": self.start_at_login,
            "show_menu_bar_icon": self.show_menu_bar_icon,
            "preview_mode": self.preview_mode,
            "confirm_destructive_actions": self.confirm_destructive_actions,
            "trash": self.trash.to_dict(),
            "notifications": self.notifications.to_dict(),
            "logging": self.logging.to_dict(),
            "watcher": self.watcher.to_dict(),
            "theme": self.theme,
            "language": self.language,
            "show_preview_on_hover": self.show_preview_on_hover,
            "max_concurrent_actions": self.max_concurrent_actions,
            "action_timeout_seconds": self.action_timeout_seconds,
            "retry_failed_actions": self.retry_failed_actions,
            "retry_count": self.retry_count,
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Settings":
        """Create settings from dictionary."""
        return cls(
            start_at_login=data.get("start_at_login", False),
            show_menu_bar_icon=data.get("show_menu_bar_icon", True),
            preview_mode=data.get("preview_mode", False),
            confirm_destructive_actions=data.get("confirm_destructive_actions", True),
            trash=TrashSettings.from_dict(data.get("trash", {})),
            notifications=NotificationSettings.from_dict(data.get("notifications", {})),
            logging=LoggingSettings.from_dict(data.get("logging", {})),
            watcher=WatcherSettings.from_dict(data.get("watcher", {})),
            theme=data.get("theme", "system"),
            language=data.get("language", "en"),
            show_preview_on_hover=data.get("show_preview_on_hover", True),
            max_concurrent_actions=data.get("max_concurrent_actions", 5),
            action_timeout_seconds=data.get("action_timeout_seconds", 300),
            retry_failed_actions=data.get("retry_failed_actions", False),
            retry_count=data.get("retry_count", 3),
        )



"""
Command-line interface for SortMeOut.

Provides a full-featured CLI for managing file automation rules.
"""

from __future__ import annotations

import json
import os
import sys
from pathlib import Path
from typing import Optional

import click
from rich.console import Console
from rich.table import Table
from rich.panel import Panel
from rich.syntax import Syntax
from rich import print as rprint

from sortmeout import SortMeOut, Rule, Condition, Action, __version__
from sortmeout.config.manager import ConfigManager
from sortmeout.utils.logger import setup_logging
from sortmeout.macos.trash import get_trash_info, empty_trash, TrashManager

console = Console()


@click.group()
@click.version_option(__version__, prog_name="SortMeOut")
@click.option("--verbose", "-v", is_flag=True, help="Enable verbose output")
@click.option("--config", "-c", type=click.Path(), help="Path to config file")
@click.pass_context
def main(ctx, verbose: bool, config: Optional[str]):
    """
    SortMeOut - Automated file organization for macOS.

    Watch folders and organize files automatically based on rules you define.
    """
    ctx.ensure_object(dict)
    ctx.obj["verbose"] = verbose
    ctx.obj["config"] = config

    if verbose:
        setup_logging(level="DEBUG")
    else:
        setup_logging(level="INFO")


@main.command()
@click.option("--preview", "-p", is_flag=True, help="Preview mode (don't execute actions)")
@click.option("--foreground", "-f", is_flag=True, help="Run in foreground")
@click.pass_context
def start(ctx, preview: bool, foreground: bool):
    """Start watching configured folders."""
    app = SortMeOut(
        config_path=ctx.obj.get("config"),
        preview_mode=preview,
        verbose=ctx.obj.get("verbose", False),
    )

    folders = app.get_folders()
    if not folders:
        console.print("[yellow]No folders configured. Use 'sortmeout folder add' to add folders.[/yellow]")
        return

    console.print(f"[green]Starting SortMeOut...[/green]")
    console.print(f"Watching {len(folders)} folder(s)")

    if preview:
        console.print("[yellow]Preview mode: Actions will not be executed[/yellow]")

    for folder in folders:
        rules = app.get_rules(folder)
        console.print(f"  • {folder} ({len(rules)} rules)")

    if foreground:
        console.print("\nPress Ctrl+C to stop...")
        try:
            app.start()
        except KeyboardInterrupt:
            console.print("\n[yellow]Stopping...[/yellow]")
            app.stop()
    else:
        thread = app.start_background()
        console.print("[green]SortMeOut running in background[/green]")


@main.command()
@click.pass_context
def stop(ctx):
    """Stop watching folders."""
    console.print("[yellow]Stopping SortMeOut...[/yellow]")
    # In a real implementation, this would communicate with a running daemon
    console.print("[green]Stopped[/green]")


@main.command()
@click.pass_context
def status(ctx):
    """Show current status."""
    app = SortMeOut(config_path=ctx.obj.get("config"))

    folders = app.get_folders()
    stats = app.get_stats()

    # Status panel
    console.print(Panel.fit(
        f"[bold]SortMeOut Status[/bold]\n\n"
        f"Folders: {len(folders)}\n"
        f"Running: {'Yes' if app.is_running else 'No'}\n"
        f"Preview Mode: {'Yes' if app.preview_mode else 'No'}",
        title="Status",
    ))

    # Folders table
    if folders:
        table = Table(title="Watched Folders")
        table.add_column("Folder", style="cyan")
        table.add_column("Rules", justify="right")

        for folder in folders:
            rules = app.get_rules(folder)
            table.add_row(folder, str(len(rules)))

        console.print(table)
    else:
        console.print("[yellow]No folders configured[/yellow]")


# Folder management commands
@main.group()
def folder():
    """Manage watched folders."""
    pass


@folder.command("add")
@click.argument("path", type=click.Path(exists=True))
@click.option("--recursive", "-r", is_flag=True, help="Watch subdirectories")
@click.pass_context
def folder_add(ctx, path: str, recursive: bool):
    """Add a folder to watch."""
    app = SortMeOut(config_path=ctx.obj.get("config"))

    try:
        if app.add_folder(path, recursive=recursive):
            console.print(f"[green]Added folder: {path}[/green]")
            if recursive:
                console.print("  (watching subdirectories)")
        else:
            console.print(f"[yellow]Folder already being watched: {path}[/yellow]")
    except ValueError as e:
        console.print(f"[red]Error: {e}[/red]")


@folder.command("remove")
@click.argument("path")
@click.pass_context
def folder_remove(ctx, path: str):
    """Remove a folder from watching."""
    app = SortMeOut(config_path=ctx.obj.get("config"))

    if app.remove_folder(path):
        console.print(f"[green]Removed folder: {path}[/green]")
    else:
        console.print(f"[red]Folder not found: {path}[/red]")


@folder.command("list")
@click.pass_context
def folder_list(ctx):
    """List all watched folders."""
    app = SortMeOut(config_path=ctx.obj.get("config"))

    folders = app.get_folders()

    if not folders:
        console.print("[yellow]No folders configured[/yellow]")
        return

    table = Table(title="Watched Folders")
    table.add_column("Path", style="cyan")
    table.add_column("Rules", justify="right")

    for folder in folders:
        rules = app.get_rules(folder)
        table.add_row(folder, str(len(rules)))

    console.print(table)


@folder.command("process")
@click.argument("path")
@click.option("--preview", "-p", is_flag=True, help="Preview only")
@click.pass_context
def folder_process(ctx, path: str, preview: bool):
    """Process all existing files in a folder."""
    app = SortMeOut(config_path=ctx.obj.get("config"), preview_mode=preview)

    try:
        result = app.process_folder(path)
        console.print(f"[green]Processed {result['processed']} files[/green]")
    except ValueError as e:
        console.print(f"[red]Error: {e}[/red]")


# Rule management commands
@main.group()
def rule():
    """Manage rules."""
    pass


@rule.command("add")
@click.argument("folder")
@click.argument("name")
@click.option("--condition", "-c", multiple=True, help="Condition (attribute:operator:value)")
@click.option("--action", "-a", multiple=True, help="Action (type:param=value)")
@click.pass_context
def rule_add(ctx, folder: str, name: str, condition: tuple, action: tuple):
    """Add a rule to a folder."""
    app = SortMeOut(config_path=ctx.obj.get("config"))

    # Parse conditions
    conditions = []
    for cond_str in condition:
        parts = cond_str.split(":", 2)
        if len(parts) >= 2:
            attr, op = parts[0], parts[1]
            value = parts[2] if len(parts) > 2 else ""
            conditions.append(Condition(attr, op, value))

    # Parse actions
    actions = []
    for action_str in action:
        if ":" in action_str:
            action_type, params_str = action_str.split(":", 1)
            params = {}
            for param in params_str.split(","):
                if "=" in param:
                    key, value = param.split("=", 1)
                    params[key.strip()] = value.strip()
            actions.append(Action(action_type, **params))
        else:
            actions.append(Action(action_str))

    try:
        rule = Rule(name=name, conditions=conditions, actions=actions)

        if app.add_rule(folder, rule):
            console.print(f"[green]Added rule '{name}' to {folder}[/green]")
        else:
            console.print(f"[yellow]Rule '{name}' already exists[/yellow]")
    except ValueError as e:
        console.print(f"[red]Error: {e}[/red]")


@rule.command("remove")
@click.argument("folder")
@click.argument("name")
@click.pass_context
def rule_remove(ctx, folder: str, name: str):
    """Remove a rule from a folder."""
    app = SortMeOut(config_path=ctx.obj.get("config"))

    if app.remove_rule(folder, name):
        console.print(f"[green]Removed rule '{name}' from {folder}[/green]")
    else:
        console.print(f"[red]Rule not found: {name}[/red]")


@rule.command("list")
@click.argument("folder")
@click.pass_context
def rule_list(ctx, folder: str):
    """List rules for a folder."""
    app = SortMeOut(config_path=ctx.obj.get("config"))

    rules = app.get_rules(folder)

    if not rules:
        console.print(f"[yellow]No rules for folder: {folder}[/yellow]")
        return

    table = Table(title=f"Rules for {folder}")
    table.add_column("#", justify="right", style="dim")
    table.add_column("Name", style="cyan")
    table.add_column("Enabled", justify="center")
    table.add_column("Conditions")
    table.add_column("Actions")

    for i, rule in enumerate(rules, 1):
        enabled = "✓" if rule.enabled else "✗"
        conditions = str(len(rule.conditions))
        actions = str(len(rule.actions))
        table.add_row(str(i), rule.name, enabled, conditions, actions)

    console.print(table)


@rule.command("show")
@click.argument("folder")
@click.argument("name")
@click.pass_context
def rule_show(ctx, folder: str, name: str):
    """Show details of a rule."""
    app = SortMeOut(config_path=ctx.obj.get("config"))

    rules = app.get_rules(folder)
    rule = next((r for r in rules if r.name == name), None)

    if not rule:
        console.print(f"[red]Rule not found: {name}[/red]")
        return

    # Rule details
    console.print(Panel.fit(
        f"[bold]{rule.name}[/bold]\n\n"
        f"Enabled: {'Yes' if rule.enabled else 'No'}\n"
        f"Match Mode: {rule.match_mode.value}\n"
        f"Continue Processing: {'Yes' if rule.continue_processing else 'No'}\n"
        f"Description: {rule.description or 'None'}",
        title="Rule Details",
    ))

    # Conditions
    if rule.conditions:
        console.print("\n[bold]Conditions:[/bold]")
        for i, cond in enumerate(rule.conditions, 1):
            console.print(f"  {i}. {cond}")

    # Actions
    if rule.actions:
        console.print("\n[bold]Actions:[/bold]")
        for i, action in enumerate(rule.actions, 1):
            console.print(f"  {i}. {action}")


@rule.command("enable")
@click.argument("folder")
@click.argument("name")
@click.pass_context
def rule_enable(ctx, folder: str, name: str):
    """Enable a rule."""
    app = SortMeOut(config_path=ctx.obj.get("config"))

    rules = app.get_rules(folder)
    rule = next((r for r in rules if r.name == name), None)

    if not rule:
        console.print(f"[red]Rule not found: {name}[/red]")
        return

    rule.enabled = True
    app.update_rule(folder, name, rule)
    console.print(f"[green]Enabled rule: {name}[/green]")


@rule.command("disable")
@click.argument("folder")
@click.argument("name")
@click.pass_context
def rule_disable(ctx, folder: str, name: str):
    """Disable a rule."""
    app = SortMeOut(config_path=ctx.obj.get("config"))

    rules = app.get_rules(folder)
    rule = next((r for r in rules if r.name == name), None)

    if not rule:
        console.print(f"[red]Rule not found: {name}[/red]")
        return

    rule.enabled = False
    app.update_rule(folder, name, rule)
    console.print(f"[yellow]Disabled rule: {name}[/yellow]")


@rule.command("export")
@click.argument("folder")
@click.argument("output", type=click.Path())
@click.pass_context
def rule_export(ctx, folder: str, output: str):
    """Export rules to a file."""
    app = SortMeOut(config_path=ctx.obj.get("config"))

    if app.export_rules(folder, output):
        console.print(f"[green]Exported rules to: {output}[/green]")
    else:
        console.print(f"[red]Failed to export rules[/red]")


@rule.command("import")
@click.argument("folder")
@click.argument("input", type=click.Path(exists=True))
@click.pass_context
def rule_import(ctx, folder: str, input: str):
    """Import rules from a file."""
    app = SortMeOut(config_path=ctx.obj.get("config"))

    count = app.import_rules(folder, input)
    console.print(f"[green]Imported {count} rules[/green]")


# Trash management commands
@main.group()
def trash():
    """Manage Trash."""
    pass


@trash.command("status")
def trash_status():
    """Show Trash status."""
    info = get_trash_info()

    console.print(Panel.fit(
        f"[bold]Trash Status[/bold]\n\n"
        f"Items: {info.item_count}\n"
        f"Size: {info.size_human}\n"
        f"Oldest: {info.oldest_item_date.strftime('%Y-%m-%d') if info.oldest_item_date else 'N/A'}\n"
        f"Newest: {info.newest_item_date.strftime('%Y-%m-%d') if info.newest_item_date else 'N/A'}",
        title="Trash",
    ))


@trash.command("empty")
@click.option("--secure", "-s", is_flag=True, help="Secure empty")
@click.confirmation_option(prompt="Are you sure you want to empty the Trash?")
def trash_empty(secure: bool):
    """Empty the Trash."""
    if empty_trash(secure=secure):
        console.print("[green]Trash emptied[/green]")
    else:
        console.print("[red]Failed to empty Trash[/red]")


@trash.command("clean")
@click.option("--days", "-d", default=30, help="Delete items older than days")
@click.option("--size", "-s", default=10.0, help="Maximum size in GB")
def trash_clean(days: int, size: float):
    """Clean Trash based on age/size limits."""
    manager = TrashManager(max_age_days=days, max_size_gb=size)
    result = manager.run_cleanup()

    deleted_count = len(result["deleted_by_age"]) + len(result["deleted_by_size"])
    console.print(f"[green]Deleted {deleted_count} items[/green]")


# Config management commands
@main.group()
def config():
    """Manage configuration."""
    pass


@config.command("show")
@click.pass_context
def config_show(ctx):
    """Show current configuration."""
    manager = ConfigManager(ctx.obj.get("config"))
    cfg = manager.load_config()

    syntax = Syntax(json.dumps(cfg, indent=2, default=str), "json")
    console.print(syntax)


@config.command("export")
@click.argument("output", type=click.Path())
@click.pass_context
def config_export(ctx, output: str):
    """Export configuration to file."""
    manager = ConfigManager(ctx.obj.get("config"))

    if manager.export_config(output):
        console.print(f"[green]Configuration exported to: {output}[/green]")
    else:
        console.print("[red]Failed to export configuration[/red]")


@config.command("import")
@click.argument("input", type=click.Path(exists=True))
@click.option("--merge", "-m", is_flag=True, help="Merge with existing config")
@click.pass_context
def config_import(ctx, input: str, merge: bool):
    """Import configuration from file."""
    manager = ConfigManager(ctx.obj.get("config"))

    if manager.import_config(input, merge=merge):
        console.print(f"[green]Configuration imported from: {input}[/green]")
    else:
        console.print("[red]Failed to import configuration[/red]")


@config.command("reset")
@click.confirmation_option(prompt="Are you sure you want to reset configuration?")
@click.pass_context
def config_reset(ctx):
    """Reset configuration to defaults."""
    manager = ConfigManager(ctx.obj.get("config"))

    if manager.reset_config():
        console.print("[green]Configuration reset to defaults[/green]")
    else:
        console.print("[red]Failed to reset configuration[/red]")


# Test command
@main.command("test")
@click.argument("file", type=click.Path(exists=True))
@click.argument("folder")
@click.option("--preview", "-p", is_flag=True, default=True, help="Preview mode")
@click.pass_context
def test_file(ctx, file: str, folder: str, preview: bool):
    """Test rules against a file."""
    app = SortMeOut(config_path=ctx.obj.get("config"), preview_mode=preview)

    from sortmeout.utils.file_info import get_file_info
    from sortmeout.core.engine import RuleEngine

    # Get file info
    file_info = get_file_info(file)
    console.print(Panel.fit(
        f"[bold]File: {os.path.basename(file)}[/bold]\n\n"
        f"Size: {file_info.get('size_human', 'N/A')}\n"
        f"Extension: {file_info.get('extension', 'N/A')}\n"
        f"Kind: {file_info.get('kind', 'N/A')}",
        title="File Info",
    ))

    # Test rules
    rules = app.get_rules(folder)
    engine = RuleEngine(preview_mode=True)

    console.print("\n[bold]Rule Matching:[/bold]")
    for rule in rules:
        matches = engine.evaluate_rule(rule, file, file_info)
        status = "[green]✓ MATCH[/green]" if matches else "[dim]✗ No match[/dim]"
        console.print(f"  {rule.name}: {status}")

        if matches:
            console.print("    [bold]Actions would execute:[/bold]")
            for action in rule.actions:
                console.print(f"      • {action}")


if __name__ == "__main__":
    main()



"""
Logging configuration for SortMeOut.
"""

from __future__ import annotations

import logging
import os
import sys
from datetime import datetime
from logging.handlers import RotatingFileHandler
from pathlib import Path
from typing import Optional

import appdirs

# Application name for directories
APP_NAME = "SortMeOut"
APP_AUTHOR = "SortMeOut"

# Default log format
DEFAULT_FORMAT = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
SIMPLE_FORMAT = "%(levelname)s: %(message)s"
DETAILED_FORMAT = "%(asctime)s - %(name)s - %(levelname)s - %(filename)s:%(lineno)d - %(message)s"


def get_log_directory() -> Path:
    """Get the log directory path."""
    log_dir = Path(appdirs.user_log_dir(APP_NAME, APP_AUTHOR))
    log_dir.mkdir(parents=True, exist_ok=True)
    return log_dir


def setup_logging(
    level: int = logging.INFO,
    log_file: Optional[str] = None,
    console: bool = True,
    file_logging: bool = True,
    max_file_size: int = 10 * 1024 * 1024,  # 10 MB
    backup_count: int = 5,
    format_string: Optional[str] = None,
) -> logging.Logger:
    """
    Set up logging for the application.

    Args:
        level: Logging level.
        log_file: Path to log file. If None, uses default location.
        console: Enable console logging.
        file_logging: Enable file logging.
        max_file_size: Maximum log file size before rotation.
        backup_count: Number of backup files to keep.
        format_string: Custom format string.

    Returns:
        Root logger instance.
    """
    # Get root logger for our package
    root_logger = logging.getLogger("sortmeout")
    root_logger.setLevel(level)

    # Clear existing handlers
    root_logger.handlers.clear()

    # Determine format
    if format_string is None:
        format_string = DETAILED_FORMAT if level == logging.DEBUG else DEFAULT_FORMAT

    formatter = logging.Formatter(format_string)

    # Console handler
    if console:
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setLevel(level)
        console_handler.setFormatter(logging.Formatter(SIMPLE_FORMAT))
        root_logger.addHandler(console_handler)

    # File handler
    if file_logging:
        if log_file is None:
            log_dir = get_log_directory()
            log_file = str(log_dir / "sortmeout.log")

        file_handler = RotatingFileHandler(
            log_file,
            maxBytes=max_file_size,
            backupCount=backup_count,
        )
        file_handler.setLevel(level)
        file_handler.setFormatter(formatter)
        root_logger.addHandler(file_handler)

    return root_logger


def get_logger(name: str) -> logging.Logger:
    """
    Get a logger for a module.

    Args:
        name: Logger name (usually __name__).

    Returns:
        Logger instance.
    """
    # Ensure name is under our package namespace
    if not name.startswith("sortmeout"):
        name = f"sortmeout.{name}"

    return logging.getLogger(name)


class ActionLogger:
    """
    Specialized logger for tracking file actions.

    Logs actions to a separate file for easy auditing.
    """

    def __init__(self, log_file: Optional[str] = None):
        """
        Initialize action logger.

        Args:
            log_file: Path to action log file.
        """
        self.logger = logging.getLogger("sortmeout.actions")
        self.logger.setLevel(logging.INFO)

        if log_file is None:
            log_dir = get_log_directory()
            log_file = str(log_dir / "actions.log")

        handler = RotatingFileHandler(
            log_file,
            maxBytes=10 * 1024 * 1024,
            backupCount=10,
        )
        handler.setFormatter(logging.Formatter(
            "%(asctime)s | %(message)s"
        ))
        self.logger.addHandler(handler)

    def log_action(
        self,
        action_type: str,
        source: str,
        destination: Optional[str] = None,
        success: bool = True,
        details: Optional[str] = None,
    ) -> None:
        """
        Log a file action.

        Args:
            action_type: Type of action performed.
            source: Source file path.
            destination: Destination path (if applicable).
            success: Whether action succeeded.
            details: Additional details.
        """
        status = "SUCCESS" if success else "FAILED"

        message_parts = [
            status,
            action_type.upper(),
            f"src={source}",
        ]

        if destination:
            message_parts.append(f"dst={destination}")

        if details:
            message_parts.append(f"({details})")

        message = " | ".join(message_parts)

        if success:
            self.logger.info(message)
        else:
            self.logger.error(message)

    def get_recent_actions(self, count: int = 100) -> list:
        """
        Get recent logged actions.

        Args:
            count: Number of recent actions to retrieve.

        Returns:
            List of action log entries.
        """
        log_dir = get_log_directory()
        log_file = log_dir / "actions.log"

        if not log_file.exists():
            return []

        with open(log_file, "r") as f:
            lines = f.readlines()

        return lines[-count:]


# Global action logger instance
_action_logger: Optional[ActionLogger] = None


def get_action_logger() -> ActionLogger:
    """Get the global action logger instance."""
    global _action_logger
    if _action_logger is None:
        _action_logger = ActionLogger()
    return _action_logger



"""
macOS Finder tags management.

Provides functions for reading and modifying Finder tags on files.
"""

from __future__ import annotations

import plistlib
import subprocess
from typing import List, Optional

from sortmeout.utils.logger import get_logger

logger = get_logger(__name__)

# Standard macOS tag colors
TAG_COLORS = {
    "none": 0,
    "gray": 1,
    "green": 2,
    "purple": 3,
    "blue": 4,
    "yellow": 5,
    "red": 6,
    "orange": 7,
}

# Reverse mapping
COLOR_NAMES = {v: k for k, v in TAG_COLORS.items()}


def get_tags(file_path: str) -> List[str]:
    """
    Get Finder tags for a file.

    Args:
        file_path: Path to the file.

    Returns:
        List of tag names.
    """
    try:
        result = subprocess.run(
            ["xattr", "-px", "com.apple.metadata:_kMDItemUserTags", file_path],
            capture_output=True,
            text=True,
            timeout=5,
        )

        if result.returncode != 0:
            return []

        # Parse hex output to binary
        hex_data = result.stdout.replace(" ", "").replace("\n", "")
        binary_data = bytes.fromhex(hex_data)

        # Parse plist
        tags_data = plistlib.loads(binary_data)

        # Extract tag names (format: "name\ncolor_index")
        tags = []
        for tag in tags_data:
            if isinstance(tag, str):
                name = tag.split("\n")[0]
                tags.append(name)

        return tags

    except Exception as e:
        logger.debug("Failed to get tags for %s: %s", file_path, e)
        return []


def set_tags(file_path: str, tags: List[str]) -> bool:
    """
    Set Finder tags for a file (replaces all existing tags).

    Args:
        file_path: Path to the file.
        tags: List of tag names to set.

    Returns:
        True if successful.
    """
    try:
        if not tags:
            # Remove all tags
            subprocess.run(
                ["xattr", "-d", "com.apple.metadata:_kMDItemUserTags", file_path],
                capture_output=True,
                timeout=5,
            )
            return True

        # Create plist data
        tag_data = [f"{tag}\n0" for tag in tags]  # 0 = no color
        plist_data = plistlib.dumps(tag_data)

        # Convert to hex string for xattr
        hex_string = plist_data.hex()

        # Write using xattr
        result = subprocess.run(
            ["xattr", "-wx", "com.apple.metadata:_kMDItemUserTags", hex_string, file_path],
            capture_output=True,
            timeout=5,
        )

        return result.returncode == 0

    except Exception as e:
        logger.error("Failed to set tags for %s: %s", file_path, e)
        return False


def add_tags(file_path: str, tags: List[str]) -> bool:
    """
    Add tags to a file (preserves existing tags).

    Args:
        file_path: Path to the file.
        tags: List of tag names to add.

    Returns:
        True if successful.
    """
    existing = get_tags(file_path)
    new_tags = list(set(existing + tags))
    return set_tags(file_path, new_tags)


def remove_tags(file_path: str, tags: List[str]) -> bool:
    """
    Remove specific tags from a file.

    Args:
        file_path: Path to the file.
        tags: List of tag names to remove.

    Returns:
        True if successful.
    """
    existing = get_tags(file_path)
    new_tags = [t for t in existing if t not in tags]
    return set_tags(file_path, new_tags)


def has_tag(file_path: str, tag: str) -> bool:
    """
    Check if a file has a specific tag.

    Args:
        file_path: Path to the file.
        tag: Tag name to check.

    Returns:
        True if file has the tag.
    """
    tags = get_tags(file_path)
    return tag in tags


def clear_tags(file_path: str) -> bool:
    """
    Remove all tags from a file.

    Args:
        file_path: Path to the file.

    Returns:
        True if successful.
    """
    return set_tags(file_path, [])


def get_all_tags() -> List[str]:
    """
    Get all tags defined in the system.

    Returns:
        List of all tag names.
    """
    try:
        # Read from Finder preferences
        result = subprocess.run(
            ["defaults", "read", "com.apple.finder", "FavoriteTagNames"],
            capture_output=True,
            text=True,
            timeout=5,
        )

        if result.returncode == 0:
            # Parse the output (it's in plist format)
            import re
            tags = re.findall(r'"([^"]+)"', result.stdout)
            return tags
    except Exception as e:
        logger.debug("Failed to get all tags: %s", e)

    # Return default tags
    return ["Red", "Orange", "Yellow", "Green", "Blue", "Purple", "Gray"]


def set_finder_comment(file_path: str, comment: str) -> bool:
    """
    Set Finder comment for a file.

    Args:
        file_path: Path to the file.
        comment: Comment to set.

    Returns:
        True if successful.
    """
    try:
        # Escape special characters
        comment = comment.replace('"', '\\"')

        script = f'''
            tell application "Finder"
                set theFile to POSIX file "{file_path}" as alias
                set comment of theFile to "{comment}"
            end tell
        '''

        result = subprocess.run(
            ["osascript", "-e", script],
            capture_output=True,
            timeout=10,
        )

        return result.returncode == 0

    except Exception as e:
        logger.error("Failed to set Finder comment: %s", e)
        return False


def get_finder_comment(file_path: str) -> Optional[str]:
    """
    Get Finder comment for a file.

    Args:
        file_path: Path to the file.

    Returns:
        Comment string or None.
    """
    try:
        script = f'''
            tell application "Finder"
                set theFile to POSIX file "{file_path}" as alias
                get comment of theFile
            end tell
        '''

        result = subprocess.run(
            ["osascript", "-e", script],
            capture_output=True,
            text=True,
            timeout=10,
        )

        if result.returncode == 0:
            return result.stdout.strip()
        return None

    except Exception as e:
        logger.debug("Failed to get Finder comment: %s", e)
        return None


def set_label_color(file_path: str, color: str | int) -> bool:
    """
    Set Finder label color for a file.

    Args:
        file_path: Path to the file.
        color: Color name or index (0-7).

    Returns:
        True if successful.
    """
    if isinstance(color, str):
        color_index = TAG_COLORS.get(color.lower(), 0)
    else:
        color_index = color

    try:
        script = f'''
            tell application "Finder"
                set theFile to POSIX file "{file_path}" as alias
                set label index of theFile to {color_index}
            end tell
        '''

        result = subprocess.run(
            ["osascript", "-e", script],
            capture_output=True,
            timeout=10,
        )

        return result.returncode == 0

    except Exception as e:
        logger.error("Failed to set label color: %s", e)
        return False



"""
Utility functions for SortMeOut.
"""

from sortmeout.utils.logger import setup_logging, get_logger
from sortmeout.utils.file_info import get_file_info

__all__ = [
    "setup_logging",
    "get_logger",
    "get_file_info",
]



"""
Condition definitions for rule matching.

Conditions define the criteria that files must meet to match a rule.
They can be combined using groups with AND, OR, NOT logic.
"""

from __future__ import annotations

import fnmatch
import os
import re
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union
import uuid


class ConditionAttribute(Enum):
    """Available attributes for conditions."""
    # File name attributes
    NAME = "name"
    EXTENSION = "extension"
    FULL_NAME = "full_name"  # name + extension

    # Path attributes
    PATH = "path"
    PARENT_FOLDER = "parent_folder"

    # Size attributes
    SIZE = "size"
    SIZE_BYTES = "size_bytes"

    # Date attributes
    DATE_CREATED = "date_created"
    DATE_MODIFIED = "date_modified"
    DATE_ACCESSED = "date_accessed"
    DATE_ADDED = "date_added"  # macOS specific

    # Type attributes
    FILE_TYPE = "file_type"  # MIME type
    KIND = "kind"  # macOS kind (e.g., "PDF document")
    UTI = "uti"  # Uniform Type Identifier

    # Content attributes
    CONTENTS = "contents"  # Text file contents

    # macOS specific
    TAGS = "tags"
    FINDER_COMMENT = "finder_comment"
    WHERE_FROM = "where_from"  # Download source URL
    SPOTLIGHT = "spotlight"  # Any Spotlight attribute

    # Custom
    CUSTOM = "custom"


class ConditionOperator(Enum):
    """Available operators for conditions."""
    # String operators
    EQUALS = "equals"
    NOT_EQUALS = "not_equals"
    CONTAINS = "contains"
    NOT_CONTAINS = "not_contains"
    STARTS_WITH = "starts_with"
    ENDS_WITH = "ends_with"
    MATCHES_REGEX = "matches_regex"
    MATCHES_GLOB = "matches_glob"

    # Numeric operators
    GREATER_THAN = "greater_than"
    LESS_THAN = "less_than"
    GREATER_OR_EQUAL = "greater_or_equal"
    LESS_OR_EQUAL = "less_or_equal"
    BETWEEN = "between"

    # Date operators
    IS_TODAY = "is_today"
    IS_YESTERDAY = "is_yesterday"
    IS_THIS_WEEK = "is_this_week"
    IS_LAST_WEEK = "is_last_week"
    IS_THIS_MONTH = "is_this_month"
    IS_LAST_MONTH = "is_last_month"
    IS_THIS_YEAR = "is_this_year"
    WITHIN_LAST = "within_last"  # e.g., within last 7 days
    NOT_WITHIN_LAST = "not_within_last"
    BEFORE = "before"
    AFTER = "after"

    # List operators
    IN_LIST = "in_list"
    NOT_IN_LIST = "not_in_list"

    # Boolean operators
    IS_TRUE = "is_true"
    IS_FALSE = "is_false"

    # Existence operators
    EXISTS = "exists"
    NOT_EXISTS = "not_exists"
    IS_EMPTY = "is_empty"
    IS_NOT_EMPTY = "is_not_empty"


class ConditionGroupMode(Enum):
    """How conditions in a group are combined."""
    ALL = "all"  # AND
    ANY = "any"  # OR
    NONE = "none"  # NOT


def parse_size(size_str: str) -> int:
    """
    Parse a size string into bytes.

    Args:
        size_str: Size string like "10MB", "1.5GB", "500KB"

    Returns:
        Size in bytes.
    """
    if isinstance(size_str, (int, float)):
        return int(size_str)

    size_str = size_str.strip().upper()

    units = {
        'B': 1,
        'KB': 1024,
        'MB': 1024 ** 2,
        'GB': 1024 ** 3,
        'TB': 1024 ** 4,
        'K': 1024,
        'M': 1024 ** 2,
        'G': 1024 ** 3,
        'T': 1024 ** 4,
    }

    # Match number followed by optional unit
    match = re.match(r'^([\d.]+)\s*([A-Z]*B?)$', size_str)
    if match:
        number = float(match.group(1))
        unit = match.group(2) or 'B'
        return int(number * units.get(unit, 1))

    return int(float(size_str))


def parse_duration(duration_str: str) -> timedelta:
    """
    Parse a duration string into timedelta.

    Args:
        duration_str: Duration string like "7 days", "2 hours", "30 minutes"

    Returns:
        timedelta object.
    """
    if isinstance(duration_str, timedelta):
        return duration_str

    duration_str = duration_str.strip().lower()

    patterns = [
        (r'(\d+)\s*(?:d|days?)', lambda m: timedelta(days=int(m.group(1)))),
        (r'(\d+)\s*(?:h|hours?)', lambda m: timedelta(hours=int(m.group(1)))),
        (r'(\d+)\s*(?:m|min(?:utes?)?)', lambda m: timedelta(minutes=int(m.group(1)))),
        (r'(\d+)\s*(?:s|sec(?:onds?)?)', lambda m: timedelta(seconds=int(m.group(1)))),
        (r'(\d+)\s*(?:w|weeks?)', lambda m: timedelta(weeks=int(m.group(1)))),
    ]

    for pattern, handler in patterns:
        match = re.match(pattern, duration_str)
        if match:
            return handler(match)

    # Default to days if no unit specified
    return timedelta(days=int(duration_str))


@dataclass
class Condition:
    """
    A single condition for file matching.

    Attributes:
        attribute: The file attribute to check.
        operator: The comparison operator.
        value: The value to compare against.
        case_sensitive: Whether string comparisons are case-sensitive.
        negate: Invert the condition result.
        id: Unique identifier.

    Example:
        >>> cond = Condition("extension", "equals", "pdf")
        >>> cond.evaluate({"extension": "pdf"})
        True
    """

    attribute: str | ConditionAttribute
    operator: str | ConditionOperator
    value: Any
    case_sensitive: bool = False
    negate: bool = False
    id: str = field(default_factory=lambda: str(uuid.uuid4()))

    def __post_init__(self) -> None:
        """Normalize attribute and operator to enum values."""
        if isinstance(self.attribute, str):
            try:
                self.attribute = ConditionAttribute(self.attribute)
            except ValueError:
                # Keep as string for custom attributes
                pass

        if isinstance(self.operator, str):
            self.operator = ConditionOperator(self.operator)

    def evaluate(self, file_info: Dict[str, Any]) -> bool:
        """
        Evaluate the condition against file information.

        Args:
            file_info: Dictionary containing file attributes.

        Returns:
            True if the condition is satisfied.
        """
        # Get the attribute value from file_info
        attr_name = self.attribute.value if isinstance(self.attribute, ConditionAttribute) else self.attribute
        actual_value = file_info.get(attr_name)

        # Handle missing attributes
        if actual_value is None:
            if self.operator in (ConditionOperator.NOT_EXISTS, ConditionOperator.IS_EMPTY):
                result = True
            elif self.operator in (ConditionOperator.EXISTS, ConditionOperator.IS_NOT_EMPTY):
                result = False
            else:
                result = False
        else:
            result = self._compare(actual_value, self.value, self.operator)

        return not result if self.negate else result

    def _compare(self, actual: Any, expected: Any, operator: ConditionOperator) -> bool:
        """
        Perform the comparison.

        Args:
            actual: The actual value from the file.
            expected: The expected value from the condition.
            operator: The comparison operator.

        Returns:
            True if the comparison succeeds.
        """
        # String operations
        if operator == ConditionOperator.EQUALS:
            return self._str_compare(actual, expected, lambda a, b: a == b)

        elif operator == ConditionOperator.NOT_EQUALS:
            return self._str_compare(actual, expected, lambda a, b: a != b)

        elif operator == ConditionOperator.CONTAINS:
            return self._str_compare(actual, expected, lambda a, b: b in a)

        elif operator == ConditionOperator.NOT_CONTAINS:
            return self._str_compare(actual, expected, lambda a, b: b not in a)

        elif operator == ConditionOperator.STARTS_WITH:
            return self._str_compare(actual, expected, lambda a, b: a.startswith(b))

        elif operator == ConditionOperator.ENDS_WITH:
            return self._str_compare(actual, expected, lambda a, b: a.endswith(b))

        elif operator == ConditionOperator.MATCHES_REGEX:
            flags = 0 if self.case_sensitive else re.IGNORECASE
            try:
                return bool(re.search(expected, str(actual), flags))
            except re.error:
                return False

        elif operator == ConditionOperator.MATCHES_GLOB:
            actual_str = str(actual)
            if not self.case_sensitive:
                actual_str = actual_str.lower()
                expected = expected.lower()
            return fnmatch.fnmatch(actual_str, expected)

        # Numeric operations
        elif operator == ConditionOperator.GREATER_THAN:
            return self._num_compare(actual, expected, lambda a, b: a > b)

        elif operator == ConditionOperator.LESS_THAN:
            return self._num_compare(actual, expected, lambda a, b: a < b)

        elif operator == ConditionOperator.GREATER_OR_EQUAL:
            return self._num_compare(actual, expected, lambda a, b: a >= b)

        elif operator == ConditionOperator.LESS_OR_EQUAL:
            return self._num_compare(actual, expected, lambda a, b: a <= b)

        elif operator == ConditionOperator.BETWEEN:
            if isinstance(expected, (list, tuple)) and len(expected) == 2:
                low, high = expected
                return self._num_compare(actual, low, lambda a, b: a >= b) and \
                       self._num_compare(actual, high, lambda a, b: a <= b)
            return False

        # Date operations
        elif operator == ConditionOperator.IS_TODAY:
            return self._is_same_day(actual, datetime.now())

        elif operator == ConditionOperator.IS_YESTERDAY:
            return self._is_same_day(actual, datetime.now() - timedelta(days=1))

        elif operator == ConditionOperator.IS_THIS_WEEK:
            return self._is_this_week(actual)

        elif operator == ConditionOperator.IS_LAST_WEEK:
            return self._is_last_week(actual)

        elif operator == ConditionOperator.IS_THIS_MONTH:
            return self._is_this_month(actual)

        elif operator == ConditionOperator.IS_LAST_MONTH:
            return self._is_last_month(actual)

        elif operator == ConditionOperator.IS_THIS_YEAR:
            return self._is_this_year(actual)

        elif operator == ConditionOperator.WITHIN_LAST:
            delta = parse_duration(expected) if isinstance(expected, str) else expected
            if isinstance(actual, datetime):
                return actual >= datetime.now() - delta
            return False

        elif operator == ConditionOperator.NOT_WITHIN_LAST:
            delta = parse_duration(expected) if isinstance(expected, str) else expected
            if isinstance(actual, datetime):
                return actual < datetime.now() - delta
            return False

        elif operator == ConditionOperator.BEFORE:
            if isinstance(actual, datetime) and isinstance(expected, datetime):
                return actual < expected
            return False

        elif operator == ConditionOperator.AFTER:
            if isinstance(actual, datetime) and isinstance(expected, datetime):
                return actual > expected
            return False

        # List operations
        elif operator == ConditionOperator.IN_LIST:
            if isinstance(expected, (list, tuple)):
                if not self.case_sensitive and isinstance(actual, str):
                    return actual.lower() in [str(e).lower() for e in expected]
                return actual in expected
            return False

        elif operator == ConditionOperator.NOT_IN_LIST:
            if isinstance(expected, (list, tuple)):
                if not self.case_sensitive and isinstance(actual, str):
                    return actual.lower() not in [str(e).lower() for e in expected]
                return actual not in expected
            return True

        # Boolean operations
        elif operator == ConditionOperator.IS_TRUE:
            return bool(actual)

        elif operator == ConditionOperator.IS_FALSE:
            return not bool(actual)

        # Existence operations
        elif operator == ConditionOperator.EXISTS:
            return actual is not None

        elif operator == ConditionOperator.NOT_EXISTS:
            return actual is None

        elif operator == ConditionOperator.IS_EMPTY:
            if actual is None:
                return True
            if isinstance(actual, str):
                return len(actual.strip()) == 0
            if hasattr(actual, '__len__'):
                return len(actual) == 0
            return False

        elif operator == ConditionOperator.IS_NOT_EMPTY:
            if actual is None:
                return False
            if isinstance(actual, str):
                return len(actual.strip()) > 0
            if hasattr(actual, '__len__'):
                return len(actual) > 0
            return True

        return False

    def _str_compare(self, actual: Any, expected: Any, comparator: Callable[[str, str], bool]) -> bool:
        """String comparison with case sensitivity handling."""
        actual_str = str(actual)
        expected_str = str(expected)

        if not self.case_sensitive:
            actual_str = actual_str.lower()
            expected_str = expected_str.lower()

        return comparator(actual_str, expected_str)

    def _num_compare(self, actual: Any, expected: Any, comparator: Callable[[float, float], bool]) -> bool:
        """Numeric comparison with size string parsing."""
        try:
            # Handle size strings
            if isinstance(expected, str) and any(c.isalpha() for c in expected):
                expected = parse_size(expected)
            if isinstance(actual, str) and any(c.isalpha() for c in actual):
                actual = parse_size(actual)

            return comparator(float(actual), float(expected))
        except (ValueError, TypeError):
            return False

    def _is_same_day(self, date: Any, target: datetime) -> bool:
        """Check if date is the same day as target."""
        if isinstance(date, datetime):
            return date.date() == target.date()
        return False

    def _is_this_week(self, date: Any) -> bool:
        """Check if date is in the current week."""
        if isinstance(date, datetime):
            now = datetime.now()
            start_of_week = now - timedelta(days=now.weekday())
            start_of_week = start_of_week.replace(hour=0, minute=0, second=0, microsecond=0)
            end_of_week = start_of_week + timedelta(days=7)
            return start_of_week <= date < end_of_week
        return False

    def _is_last_week(self, date: Any) -> bool:
        """Check if date is in the previous week."""
        if isinstance(date, datetime):
            now = datetime.now()
            start_of_this_week = now - timedelta(days=now.weekday())
            start_of_this_week = start_of_this_week.replace(hour=0, minute=0, second=0, microsecond=0)
            start_of_last_week = start_of_this_week - timedelta(days=7)
            return start_of_last_week <= date < start_of_this_week
        return False

    def _is_this_month(self, date: Any) -> bool:
        """Check if date is in the current month."""
        if isinstance(date, datetime):
            now = datetime.now()
            return date.year == now.year and date.month == now.month
        return False

    def _is_last_month(self, date: Any) -> bool:
        """Check if date is in the previous month."""
        if isinstance(date, datetime):
            now = datetime.now()
            last_month = now.replace(day=1) - timedelta(days=1)
            return date.year == last_month.year and date.month == last_month.month
        return False

    def _is_this_year(self, date: Any) -> bool:
        """Check if date is in the current year."""
        if isinstance(date, datetime):
            return date.year == datetime.now().year
        return False

    def duplicate(self) -> "Condition":
        """Create a copy of this condition."""
        return Condition(
            attribute=self.attribute,
            operator=self.operator,
            value=self.value,
            case_sensitive=self.case_sensitive,
            negate=self.negate,
        )

    def to_dict(self) -> Dict[str, Any]:
        """Convert condition to dictionary for serialization."""
        attr = self.attribute.value if isinstance(self.attribute, ConditionAttribute) else self.attribute
        op = self.operator.value if isinstance(self.operator, ConditionOperator) else self.operator

        return {
            "id": self.id,
            "attribute": attr,
            "operator": op,
            "value": self.value,
            "case_sensitive": self.case_sensitive,
            "negate": self.negate,
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Condition":
        """Create a condition from a dictionary."""
        return cls(
            id=data.get("id", str(uuid.uuid4())),
            attribute=data["attribute"],
            operator=data["operator"],
            value=data["value"],
            case_sensitive=data.get("case_sensitive", False),
            negate=data.get("negate", False),
        )

    def __str__(self) -> str:
        """Human-readable representation."""
        attr = self.attribute.value if isinstance(self.attribute, ConditionAttribute) else self.attribute
        op = self.operator.value if isinstance(self.operator, ConditionOperator) else self.operator
        neg = "NOT " if self.negate else ""
        return f"{neg}{attr} {op} {self.value!r}"


@dataclass
class ConditionGroup:
    """
    A group of conditions combined with AND, OR, or NOT logic.

    Condition groups allow for complex nested logic in rules.

    Attributes:
        conditions: List of conditions or nested groups.
        mode: How conditions are combined (all, any, none).
        id: Unique identifier.

    Example:
        >>> group = ConditionGroup(
        ...     conditions=[
        ...         Condition("extension", "equals", "pdf"),
        ...         Condition("size", "greater_than", "1MB"),
        ...     ],
        ...     mode=ConditionGroupMode.ALL
        ... )
    """

    conditions: List[Condition | "ConditionGroup"]
    mode: ConditionGroupMode = ConditionGroupMode.ALL
    id: str = field(default_factory=lambda: str(uuid.uuid4()))

    def __post_init__(self) -> None:
        """Normalize mode to enum."""
        if isinstance(self.mode, str):
            self.mode = ConditionGroupMode(self.mode)

    def evaluate(self, file_info: Dict[str, Any]) -> bool:
        """
        Evaluate the condition group against file information.

        Args:
            file_info: Dictionary containing file attributes.

        Returns:
            True if the condition group is satisfied.
        """
        if not self.conditions:
            return True

        results = [cond.evaluate(file_info) for cond in self.conditions]

        if self.mode == ConditionGroupMode.ALL:
            return all(results)
        elif self.mode == ConditionGroupMode.ANY:
            return any(results)
        elif self.mode == ConditionGroupMode.NONE:
            return not any(results)

        return False

    def add_condition(self, condition: Condition | "ConditionGroup") -> "ConditionGroup":
        """Add a condition to the group."""
        self.conditions.append(condition)
        return self

    def duplicate(self) -> "ConditionGroup":
        """Create a copy of this condition group."""
        return ConditionGroup(
            conditions=[c.duplicate() for c in self.conditions],
            mode=self.mode,
        )

    def to_dict(self) -> Dict[str, Any]:
        """Convert condition group to dictionary for serialization."""
        return {
            "id": self.id,
            "mode": self.mode.value,
            "conditions": [c.to_dict() for c in self.conditions],
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "ConditionGroup":
        """Create a condition group from a dictionary."""
        conditions = []
        for cond_data in data.get("conditions", []):
            if "conditions" in cond_data:
                conditions.append(cls.from_dict(cond_data))
            else:
                conditions.append(Condition.from_dict(cond_data))

        return cls(
            id=data.get("id", str(uuid.uuid4())),
            mode=ConditionGroupMode(data.get("mode", "all")),
            conditions=conditions,
        )

    def __str__(self) -> str:
        """Human-readable representation."""
        mode_str = self.mode.value.upper()
        conditions_str = f" {mode_str} ".join(str(c) for c in self.conditions)
        return f"({conditions_str})"


# Convenience functions for creating common conditions

def name_equals(value: str, case_sensitive: bool = False) -> Condition:
    """Create a condition that checks if file name equals value."""
    return Condition("name", "equals", value, case_sensitive=case_sensitive)


def name_contains(value: str, case_sensitive: bool = False) -> Condition:
    """Create a condition that checks if file name contains value."""
    return Condition("name", "contains", value, case_sensitive=case_sensitive)


def extension_is(ext: str) -> Condition:
    """Create a condition that checks file extension."""
    # Remove leading dot if present
    ext = ext.lstrip(".")
    return Condition("extension", "equals", ext, case_sensitive=False)


def extension_in(extensions: List[str]) -> Condition:
    """Create a condition that checks if extension is in list."""
    # Remove leading dots
    extensions = [e.lstrip(".") for e in extensions]
    return Condition("extension", "in_list", extensions, case_sensitive=False)


def size_greater_than(size: str | int) -> Condition:
    """Create a condition that checks if file size is greater than value."""
    return Condition("size", "greater_than", size)


def size_less_than(size: str | int) -> Condition:
    """Create a condition that checks if file size is less than value."""
    return Condition("size", "less_than", size)


def modified_within(duration: str) -> Condition:
    """Create a condition that checks if file was modified within duration."""
    return Condition("date_modified", "within_last", duration)


def created_within(duration: str) -> Condition:
    """Create a condition that checks if file was created within duration."""
    return Condition("date_created", "within_last", duration)


def has_tag(tag: str) -> Condition:
    """Create a condition that checks if file has a specific tag."""
    return Condition("tags", "contains", tag)


def from_url(url_pattern: str) -> Condition:
    """Create a condition that checks download source URL."""
    return Condition("where_from", "contains", url_pattern)



"""
SortMeOut - Open-source file automation and organization tool for macOS.

A powerful, rule-based file organization system inspired by Noodlesoft Hazel.
"""

__version__ = "1.0.0"
__author__ = "SortMeOut Contributors"
__license__ = "MIT"

from sortmeout.app import SortMeOut
from sortmeout.core.rule import Rule
from sortmeout.core.condition import Condition, ConditionGroup
from sortmeout.core.action import Action

__all__ = [
    "SortMeOut",
    "Rule",
    "Condition",
    "ConditionGroup",
    "Action",
    "__version__",
]



"""
macOS Trash management.

Provides functions for managing the Trash and implementing App Sweep functionality.
"""

from __future__ import annotations

import os
import subprocess
from dataclasses import dataclass
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, Dict, List, Optional

from sortmeout.utils.logger import get_logger

logger = get_logger(__name__)


@dataclass
class TrashItem:
    """Information about an item in the Trash."""
    path: str
    original_path: str
    name: str
    size: int
    deleted_date: datetime
    kind: str

    @property
    def age_days(self) -> int:
        """Days since item was deleted."""
        return (datetime.now() - self.deleted_date).days


@dataclass
class TrashInfo:
    """Overall Trash statistics."""
    item_count: int
    total_size: int
    oldest_item_date: Optional[datetime]
    newest_item_date: Optional[datetime]
    items: List[TrashItem]

    @property
    def size_human(self) -> str:
        """Human-readable size."""
        size = self.total_size
        for unit in ["B", "KB", "MB", "GB", "TB"]:
            if size < 1024:
                return f"{size:.1f} {unit}"
            size /= 1024
        return f"{size:.1f} PB"


def get_trash_path() -> Path:
    """Get the user's Trash directory path."""
    return Path.home() / ".Trash"


def get_trash_info() -> TrashInfo:
    """
    Get information about the Trash contents.

    Returns:
        TrashInfo with statistics and item list.
    """
    trash_path = get_trash_path()
    items = []
    total_size = 0
    oldest_date = None
    newest_date = None

    if not trash_path.exists():
        return TrashInfo(
            item_count=0,
            total_size=0,
            oldest_item_date=None,
            newest_item_date=None,
            items=[],
        )

    for entry in trash_path.iterdir():
        try:
            stat = entry.stat()

            # Get size (recursively for directories)
            if entry.is_dir():
                size = sum(f.stat().st_size for f in entry.rglob("*") if f.is_file())
            else:
                size = stat.st_size

            total_size += size

            # Get deletion date from .DS_Store or use modification time
            deleted_date = datetime.fromtimestamp(stat.st_mtime)

            # Update date range
            if oldest_date is None or deleted_date < oldest_date:
                oldest_date = deleted_date
            if newest_date is None or deleted_date > newest_date:
                newest_date = deleted_date

            # Determine kind
            kind = "Folder" if entry.is_dir() else "File"

            items.append(TrashItem(
                path=str(entry),
                original_path="",  # Would need .Trashes metadata
                name=entry.name,
                size=size,
                deleted_date=deleted_date,
                kind=kind,
            ))

        except (PermissionError, OSError) as e:
            logger.debug("Error reading trash item %s: %s", entry, e)
            continue

    return TrashInfo(
        item_count=len(items),
        total_size=total_size,
        oldest_item_date=oldest_date,
        newest_item_date=newest_date,
        items=items,
    )


def empty_trash(secure: bool = False) -> bool:
    """
    Empty the Trash.

    Args:
        secure: If True, securely delete files (slower).

    Returns:
        True if successful.
    """
    try:
        if secure:
            script = '''
                tell application "Finder"
                    empty trash with security
                end tell
            '''
        else:
            script = '''
                tell application "Finder"
                    empty trash
                end tell
            '''

        result = subprocess.run(
            ["osascript", "-e", script],
            capture_output=True,
            timeout=60,
        )

        if result.returncode == 0:
            logger.info("Trash emptied")
            return True
        else:
            logger.error("Failed to empty trash: %s", result.stderr)
            return False

    except Exception as e:
        logger.error("Error emptying trash: %s", e)
        return False


def delete_old_trash_items(max_age_days: int) -> List[str]:
    """
    Delete items from Trash older than specified days.

    Args:
        max_age_days: Maximum age in days.

    Returns:
        List of deleted item names.
    """
    trash_info = get_trash_info()
    cutoff = datetime.now() - timedelta(days=max_age_days)
    deleted = []

    for item in trash_info.items:
        if item.deleted_date < cutoff:
            try:
                path = Path(item.path)
                if path.is_dir():
                    import shutil
                    shutil.rmtree(path)
                else:
                    path.unlink()
                deleted.append(item.name)
                logger.info("Deleted old trash item: %s", item.name)
            except Exception as e:
                logger.error("Failed to delete %s: %s", item.name, e)

    return deleted


def trim_trash_to_size(max_size_bytes: int) -> List[str]:
    """
    Trim Trash to maximum size by removing oldest items first.

    Args:
        max_size_bytes: Maximum total size in bytes.

    Returns:
        List of deleted item names.
    """
    trash_info = get_trash_info()

    if trash_info.total_size <= max_size_bytes:
        return []

    # Sort by date (oldest first)
    items_by_date = sorted(trash_info.items, key=lambda x: x.deleted_date)

    deleted = []
    current_size = trash_info.total_size

    for item in items_by_date:
        if current_size <= max_size_bytes:
            break

        try:
            path = Path(item.path)
            if path.is_dir():
                import shutil
                shutil.rmtree(path)
            else:
                path.unlink()

            current_size -= item.size
            deleted.append(item.name)
            logger.info("Deleted trash item to free space: %s", item.name)

        except Exception as e:
            logger.error("Failed to delete %s: %s", item.name, e)

    return deleted


class TrashManager:
    """
    Automatic trash management.

    Monitors and manages the Trash based on configured policies.
    """

    def __init__(
        self,
        max_age_days: int = 30,
        max_size_gb: float = 10.0,
        enabled: bool = True,
    ):
        """
        Initialize trash manager.

        Args:
            max_age_days: Maximum age for items.
            max_size_gb: Maximum total size in GB.
            enabled: Whether management is enabled.
        """
        self.max_age_days = max_age_days
        self.max_size_gb = max_size_gb
        self.max_size_bytes = int(max_size_gb * 1024 * 1024 * 1024)
        self.enabled = enabled

    def run_cleanup(self) -> Dict[str, Any]:
        """
        Run trash cleanup based on policies.

        Returns:
            Cleanup results.
        """
        if not self.enabled:
            return {"enabled": False, "deleted": []}

        results = {
            "deleted_by_age": [],
            "deleted_by_size": [],
            "before_size": 0,
            "after_size": 0,
        }

        # Get initial state
        before = get_trash_info()
        results["before_size"] = before.total_size

        # Delete old items
        if self.max_age_days > 0:
            results["deleted_by_age"] = delete_old_trash_items(self.max_age_days)

        # Trim to size
        if self.max_size_bytes > 0:
            results["deleted_by_size"] = trim_trash_to_size(self.max_size_bytes)

        # Get final state
        after = get_trash_info()
        results["after_size"] = after.total_size

        return results

    def get_status(self) -> Dict[str, Any]:
        """
        Get current trash status.

        Returns:
            Status information.
        """
        info = get_trash_info()

        return {
            "item_count": info.item_count,
            "total_size": info.total_size,
            "size_human": info.size_human,
            "oldest_item_age_days": (datetime.now() - info.oldest_item_date).days if info.oldest_item_date else 0,
            "over_size_limit": info.total_size > self.max_size_bytes,
            "has_old_items": info.oldest_item_date and (datetime.now() - info.oldest_item_date).days > self.max_age_days,
        }


# App Sweep functionality

def find_app_support_files(app_name: str) -> List[str]:
    """
    Find support files for an application.

    Args:
        app_name: Application name or bundle identifier.

    Returns:
        List of support file paths.
    """
    support_dirs = [
        Path.home() / "Library" / "Application Support",
        Path.home() / "Library" / "Preferences",
        Path.home() / "Library" / "Caches",
        Path.home() / "Library" / "Containers",
        Path.home() / "Library" / "Logs",
        Path.home() / "Library" / "Saved Application State",
        Path("/Library/Application Support"),
        Path("/Library/Preferences"),
    ]

    found_files = []

    # Normalize app name for matching
    app_name_lower = app_name.lower()
    # Remove .app extension if present
    if app_name_lower.endswith(".app"):
        app_name_lower = app_name_lower[:-4]

    for support_dir in support_dirs:
        if not support_dir.exists():
            continue

        try:
            for entry in support_dir.iterdir():
                entry_name = entry.name.lower()

                # Match by name
                if app_name_lower in entry_name:
                    found_files.append(str(entry))
                    continue

                # Match common patterns
                if any(pattern in entry_name for pattern in [
                    app_name_lower.replace(" ", ""),
                    app_name_lower.replace(" ", "-"),
                    app_name_lower.replace(" ", "_"),
                ]):
                    found_files.append(str(entry))

        except PermissionError:
            continue

    return found_files


def get_app_support_size(app_name: str) -> int:
    """
    Get total size of support files for an application.

    Args:
        app_name: Application name.

    Returns:
        Total size in bytes.
    """
    files = find_app_support_files(app_name)
    total = 0

    for file_path in files:
        path = Path(file_path)
        try:
            if path.is_dir():
                total += sum(f.stat().st_size for f in path.rglob("*") if f.is_file())
            else:
                total += path.stat().st_size
        except (PermissionError, OSError):
            continue

    return total


def clean_app_support_files(app_name: str, to_trash: bool = True) -> List[str]:
    """
    Clean up support files for an application.

    Args:
        app_name: Application name.
        to_trash: Move to trash instead of deleting permanently.

    Returns:
        List of cleaned files.
    """
    files = find_app_support_files(app_name)
    cleaned = []

    for file_path in files:
        try:
            if to_trash:
                # Move to trash using Finder
                script = f'''
                    tell application "Finder"
                        delete POSIX file "{file_path}"
                    end tell
                '''
                subprocess.run(["osascript", "-e", script], capture_output=True, timeout=10)
            else:
                # Delete directly
                path = Path(file_path)
                if path.is_dir():
                    import shutil
                    shutil.rmtree(path)
                else:
                    path.unlink()

            cleaned.append(file_path)
            logger.info("Cleaned app support file: %s", file_path)

        except Exception as e:
            logger.error("Failed to clean %s: %s", file_path, e)

    return cleaned



"""
Configuration file management.

Handles reading and writing configuration files for SortMeOut.
"""

from __future__ import annotations

import json
import os
import shutil
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Optional

import yaml
import appdirs

from sortmeout.config.settings import Settings
from sortmeout.utils.logger import get_logger

logger = get_logger(__name__)

# Application identifiers
APP_NAME = "SortMeOut"
APP_AUTHOR = "SortMeOut"


def get_config_directory() -> Path:
    """Get the configuration directory path."""
    config_dir = Path(appdirs.user_config_dir(APP_NAME, APP_AUTHOR))
    config_dir.mkdir(parents=True, exist_ok=True)
    return config_dir


def get_data_directory() -> Path:
    """Get the data directory path."""
    data_dir = Path(appdirs.user_data_dir(APP_NAME, APP_AUTHOR))
    data_dir.mkdir(parents=True, exist_ok=True)
    return data_dir


class ConfigManager:
    """
    Manages configuration file operations.

    Handles loading, saving, and migrating configuration files.
    Supports both YAML and JSON formats.
    """

    def __init__(self, config_path: Optional[str] = None):
        """
        Initialize configuration manager.

        Args:
            config_path: Path to configuration file. If None, uses default.
        """
        if config_path:
            self.config_path = Path(config_path)
        else:
            self.config_path = get_config_directory() / "config.yaml"

        self.config_dir = self.config_path.parent
        self.config_dir.mkdir(parents=True, exist_ok=True)

        # Backup directory
        self.backup_dir = self.config_dir / "backups"
        self.backup_dir.mkdir(exist_ok=True)

        logger.debug("Config path: %s", self.config_path)

    def load_config(self) -> Dict[str, Any]:
        """
        Load configuration from file.

        Returns:
            Configuration dictionary.
        """
        if not self.config_path.exists():
            logger.info("No config file found, using defaults")
            return self._get_default_config()

        try:
            with open(self.config_path, "r") as f:
                if self.config_path.suffix in (".yaml", ".yml"):
                    config = yaml.safe_load(f) or {}
                else:
                    config = json.load(f)

            logger.info("Loaded config from: %s", self.config_path)
            return config

        except Exception as e:
            logger.error("Failed to load config: %s", e)
            return self._get_default_config()

    def save_config(self, config: Dict[str, Any]) -> bool:
        """
        Save configuration to file.

        Args:
            config: Configuration dictionary to save.

        Returns:
            True if save was successful.
        """
        try:
            # Create backup before saving
            if self.config_path.exists():
                self._create_backup()

            # Add metadata
            config["_metadata"] = {
                "version": "1.0",
                "updated_at": datetime.now().isoformat(),
                "app_version": "0.1.0",
            }

            with open(self.config_path, "w") as f:
                if self.config_path.suffix in (".yaml", ".yml"):
                    yaml.dump(config, f, default_flow_style=False, sort_keys=False)
                else:
                    json.dump(config, f, indent=2)

            logger.info("Saved config to: %s", self.config_path)
            return True

        except Exception as e:
            logger.error("Failed to save config: %s", e)
            return False

    def load_settings(self) -> Settings:
        """
        Load application settings.

        Returns:
            Settings object.
        """
        settings_path = self.config_dir / "settings.yaml"

        if settings_path.exists():
            try:
                with open(settings_path, "r") as f:
                    data = yaml.safe_load(f) or {}
                return Settings.from_dict(data)
            except Exception as e:
                logger.error("Failed to load settings: %s", e)

        return Settings()

    def save_settings(self, settings: Settings) -> bool:
        """
        Save application settings.

        Args:
            settings: Settings object to save.

        Returns:
            True if save was successful.
        """
        settings_path = self.config_dir / "settings.yaml"

        try:
            with open(settings_path, "w") as f:
                yaml.dump(settings.to_dict(), f, default_flow_style=False)
            return True
        except Exception as e:
            logger.error("Failed to save settings: %s", e)
            return False

    def _get_default_config(self) -> Dict[str, Any]:
        """Get default configuration."""
        return {
            "folders": [],
            "_metadata": {
                "version": "1.0",
                "created_at": datetime.now().isoformat(),
                "app_version": "0.1.0",
            },
        }

    def _create_backup(self) -> None:
        """Create a backup of the current config file."""
        if not self.config_path.exists():
            return

        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        backup_name = f"config_backup_{timestamp}{self.config_path.suffix}"
        backup_path = self.backup_dir / backup_name

        shutil.copy2(self.config_path, backup_path)
        logger.debug("Created config backup: %s", backup_path)

        # Keep only last 10 backups
        self._cleanup_old_backups()

    def _cleanup_old_backups(self, keep: int = 10) -> None:
        """Remove old backup files."""
        backups = sorted(self.backup_dir.glob("config_backup_*"))

        if len(backups) > keep:
            for backup in backups[:-keep]:
                backup.unlink()
                logger.debug("Removed old backup: %s", backup)

    def restore_backup(self, backup_name: Optional[str] = None) -> bool:
        """
        Restore configuration from a backup.

        Args:
            backup_name: Name of backup file. If None, uses most recent.

        Returns:
            True if restore was successful.
        """
        if backup_name:
            backup_path = self.backup_dir / backup_name
        else:
            backups = sorted(self.backup_dir.glob("config_backup_*"))
            if not backups:
                logger.warning("No backups found")
                return False
            backup_path = backups[-1]

        if not backup_path.exists():
            logger.error("Backup not found: %s", backup_path)
            return False

        try:
            shutil.copy2(backup_path, self.config_path)
            logger.info("Restored config from: %s", backup_path)
            return True
        except Exception as e:
            logger.error("Failed to restore backup: %s", e)
            return False

    def list_backups(self) -> list:
        """List available backup files."""
        return sorted([b.name for b in self.backup_dir.glob("config_backup_*")])

    def export_config(self, output_path: str, format: str = "yaml") -> bool:
        """
        Export configuration to a file.

        Args:
            output_path: Path to export to.
            format: Export format (yaml or json).

        Returns:
            True if export was successful.
        """
        config = self.load_config()

        try:
            with open(output_path, "w") as f:
                if format == "yaml":
                    yaml.dump(config, f, default_flow_style=False)
                else:
                    json.dump(config, f, indent=2)

            logger.info("Exported config to: %s", output_path)
            return True

        except Exception as e:
            logger.error("Failed to export config: %s", e)
            return False

    def import_config(self, input_path: str, merge: bool = False) -> bool:
        """
        Import configuration from a file.

        Args:
            input_path: Path to import from.
            merge: Merge with existing config instead of replacing.

        Returns:
            True if import was successful.
        """
        try:
            with open(input_path, "r") as f:
                if input_path.endswith((".yaml", ".yml")):
                    imported = yaml.safe_load(f) or {}
                else:
                    imported = json.load(f)

            if merge:
                current = self.load_config()
                # Merge folders
                current_folders = {f["path"]: f for f in current.get("folders", [])}
                for folder in imported.get("folders", []):
                    current_folders[folder["path"]] = folder
                current["folders"] = list(current_folders.values())
                config = current
            else:
                config = imported

            return self.save_config(config)

        except Exception as e:
            logger.error("Failed to import config: %s", e)
            return False

    def reset_config(self) -> bool:
        """
        Reset configuration to defaults.

        Creates a backup before resetting.

        Returns:
            True if reset was successful.
        """
        self._create_backup()
        return self.save_config(self._get_default_config())



"""
Rule execution engine.

The engine processes files against rules, evaluates conditions,
and executes actions.
"""

from __future__ import annotations

import os
import mimetypes
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional
import logging

from sortmeout.core.rule import Rule
from sortmeout.core.action import Action, ActionResult
from sortmeout.utils.logger import get_logger
from sortmeout.utils.file_info import get_file_info

logger = get_logger(__name__)


@dataclass
class ProcessingResult:
    """
    Result of processing a file through rules.

    Attributes:
        file_path: Path to the processed file.
        matched_rules: Names of rules that matched.
        executed_actions: Results of executed actions.
        errors: Any errors that occurred.
        processing_time: Time taken to process.
    """
    file_path: str
    matched_rules: List[str] = field(default_factory=list)
    executed_actions: List[ActionResult] = field(default_factory=list)
    errors: List[str] = field(default_factory=list)
    processing_time: float = 0.0
    stopped: bool = False  # Processing was stopped by a rule

    @property
    def success(self) -> bool:
        """Check if processing was successful (no errors)."""
        return len(self.errors) == 0

    def __str__(self) -> str:
        return (
            f"ProcessingResult(file={Path(self.file_path).name}, "
            f"matched={len(self.matched_rules)}, "
            f"actions={len(self.executed_actions)}, "
            f"errors={len(self.errors)})"
        )


class RuleEngine:
    """
    Engine for processing files against rules.

    The engine:
    1. Gathers file information
    2. Evaluates each rule's conditions
    3. Executes actions for matching rules
    4. Handles errors and continues/stops as configured

    Attributes:
        preview_mode: If True, actions are not actually executed.
        stop_on_error: If True, stop processing on first error.
    """

    def __init__(
        self,
        preview_mode: bool = False,
        stop_on_error: bool = False,
        max_rules_per_file: int = 100,
    ):
        """
        Initialize the rule engine.

        Args:
            preview_mode: Don't actually execute actions.
            stop_on_error: Stop on first error.
            max_rules_per_file: Maximum rules to process per file (safety limit).
        """
        self.preview_mode = preview_mode
        self.stop_on_error = stop_on_error
        self.max_rules_per_file = max_rules_per_file

        # Statistics
        self._stats = {
            "files_processed": 0,
            "rules_evaluated": 0,
            "rules_matched": 0,
            "actions_executed": 0,
            "errors": 0,
        }

    def process_file(
        self,
        file_path: str,
        rules: List[Rule],
        file_info: Optional[Dict[str, Any]] = None,
    ) -> ProcessingResult:
        """
        Process a file against a list of rules.

        Args:
            file_path: Path to the file to process.
            rules: List of rules to evaluate.
            file_info: Pre-computed file information (optional).

        Returns:
            ProcessingResult with matched rules and action results.
        """
        start_time = datetime.now()
        result = ProcessingResult(file_path=file_path)

        # Check if file still exists
        if not os.path.exists(file_path):
            result.errors.append(f"File does not exist: {file_path}")
            return result

        # Get file information
        if file_info is None:
            try:
                file_info = get_file_info(file_path)
            except Exception as e:
                result.errors.append(f"Failed to get file info: {e}")
                return result

        logger.debug("Processing file: %s", file_path)
        logger.debug("File info: %s", file_info)

        # Track current file path (may change during processing)
        current_path = file_path
        rules_processed = 0

        # Process each rule
        for rule in rules:
            if rules_processed >= self.max_rules_per_file:
                logger.warning("Max rules limit reached for file: %s", file_path)
                break

            rules_processed += 1
            self._stats["rules_evaluated"] += 1

            # Skip disabled rules
            if not rule.enabled:
                continue

            # Check if rule matches
            try:
                matches = rule.matches(file_info)
            except Exception as e:
                logger.error("Error evaluating rule '%s': %s", rule.name, e)
                result.errors.append(f"Rule '{rule.name}' evaluation error: {e}")
                if self.stop_on_error:
                    break
                continue

            if not matches:
                continue

            # Rule matched!
            logger.info("Rule '%s' matched file: %s", rule.name, current_path)
            result.matched_rules.append(rule.name)
            self._stats["rules_matched"] += 1

            # Execute actions
            action_results, new_path, stop = self._execute_actions(
                rule.actions,
                current_path,
                file_info,
            )

            result.executed_actions.extend(action_results)

            # Check for action errors
            for ar in action_results:
                if not ar.success:
                    result.errors.append(f"Action failed: {ar.error}")
                    self._stats["errors"] += 1

            # Update current path if file was moved/renamed
            if new_path and new_path != current_path:
                current_path = new_path
                # Update file_info for subsequent rules
                try:
                    file_info = get_file_info(current_path)
                except:
                    pass

            # Check if we should stop processing
            if stop or not rule.continue_processing:
                result.stopped = True
                break

            # Stop on error if configured
            if result.errors and self.stop_on_error:
                break

        # Calculate processing time
        result.processing_time = (datetime.now() - start_time).total_seconds()
        self._stats["files_processed"] += 1

        logger.debug("Finished processing: %s", result)
        return result

    def _execute_actions(
        self,
        actions: List[Action],
        file_path: str,
        file_info: Dict[str, Any],
    ) -> tuple[List[ActionResult], Optional[str], bool]:
        """
        Execute a list of actions on a file.

        Args:
            actions: Actions to execute.
            file_path: Current file path.
            file_info: File information.

        Returns:
            Tuple of (action_results, new_path, should_stop).
        """
        results = []
        current_path = file_path
        should_stop = False

        for action in actions:
            if not action.enabled:
                continue

            # Check if file still exists (may have been deleted by previous action)
            if not os.path.exists(current_path):
                logger.warning("File no longer exists, stopping actions: %s", current_path)
                break

            # Execute the action
            try:
                result = action.execute(
                    current_path,
                    file_info,
                    preview=self.preview_mode,
                )
                results.append(result)
                self._stats["actions_executed"] += 1

                # Update path if action moved/renamed the file
                if result.success and result.destination_path:
                    current_path = result.destination_path

                # Stop on error if action is configured to do so
                if not result.success and action.stop_on_error:
                    logger.warning("Action failed, stopping: %s", result.error)
                    should_stop = True
                    break

            except Exception as e:
                logger.error("Action execution error: %s", e)
                results.append(ActionResult(
                    success=False,
                    action_type=action.action_type,
                    source_path=current_path,
                    error=str(e),
                ))

                if action.stop_on_error:
                    should_stop = True
                    break

        return results, current_path if current_path != file_path else None, should_stop

    def evaluate_rule(
        self,
        rule: Rule,
        file_path: str,
        file_info: Optional[Dict[str, Any]] = None,
    ) -> bool:
        """
        Evaluate a single rule against a file without executing actions.

        Args:
            rule: Rule to evaluate.
            file_path: Path to the file.
            file_info: Pre-computed file information.

        Returns:
            True if the rule matches.
        """
        if file_info is None:
            file_info = get_file_info(file_path)

        return rule.matches(file_info)

    def preview_rule(
        self,
        rule: Rule,
        file_path: str,
        file_info: Optional[Dict[str, Any]] = None,
    ) -> List[ActionResult]:
        """
        Preview what actions a rule would perform on a file.

        Args:
            rule: Rule to preview.
            file_path: Path to the file.
            file_info: Pre-computed file information.

        Returns:
            List of ActionResults (with preview=True).
        """
        if file_info is None:
            file_info = get_file_info(file_path)

        if not rule.matches(file_info):
            return []

        results = []
        for action in rule.actions:
            if action.enabled:
                result = action.execute(file_path, file_info, preview=True)
                results.append(result)

        return results

    def get_stats(self) -> Dict[str, int]:
        """Get engine statistics."""
        return self._stats.copy()

    def reset_stats(self) -> None:
        """Reset engine statistics."""
        self._stats = {
            "files_processed": 0,
            "rules_evaluated": 0,
            "rules_matched": 0,
            "actions_executed": 0,
            "errors": 0,
        }


class BatchProcessor:
    """
    Process multiple files in batch.

    Useful for processing existing files in a folder when rules are first added.
    """

    def __init__(self, engine: RuleEngine):
        """
        Initialize batch processor.

        Args:
            engine: Rule engine to use for processing.
        """
        self.engine = engine

    def process_folder(
        self,
        folder_path: str,
        rules: List[Rule],
        recursive: bool = False,
        file_filter: Optional[callable] = None,
    ) -> List[ProcessingResult]:
        """
        Process all files in a folder.

        Args:
            folder_path: Path to the folder.
            rules: Rules to apply.
            recursive: Process subdirectories.
            file_filter: Optional function to filter files.

        Returns:
            List of ProcessingResults.
        """
        results = []
        folder = Path(folder_path)

        if recursive:
            files = folder.rglob("*")
        else:
            files = folder.glob("*")

        for path in files:
            if not path.is_file():
                continue

            if file_filter and not file_filter(str(path)):
                continue

            result = self.engine.process_file(str(path), rules)
            results.append(result)

        return results

    def process_files(
        self,
        file_paths: List[str],
        rules: List[Rule],
    ) -> List[ProcessingResult]:
        """
        Process a list of specific files.

        Args:
            file_paths: Paths to files.
            rules: Rules to apply.

        Returns:
            List of ProcessingResults.
        """
        results = []

        for file_path in file_paths:
            result = self.engine.process_file(file_path, rules)
            results.append(result)

        return results



"""
macOS-specific integrations for SortMeOut.
"""

from sortmeout.macos.tags import get_tags, set_tags, add_tags, remove_tags
from sortmeout.macos.spotlight import search_spotlight, get_metadata
from sortmeout.macos.trash import get_trash_info, empty_trash, TrashManager

__all__ = [
    "get_tags",
    "set_tags",
    "add_tags",
    "remove_tags",
    "search_spotlight",
    "get_metadata",
    "get_trash_info",
    "empty_trash",
    "TrashManager",
]
